#!/usr/bin/env python3
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

# ....................{ IMPORTS                            }....................
import numpy as np
import numpy.ma as ma
from scipy import interpolate as interp
from scipy.ndimage import gaussian_filter
# from betse.science.math import toolbox as tb
from betse.exceptions import BetseSimUnstableException
from betse.science.math import finitediff as fd

# ....................{ UTILITIES                          }....................
# Toolbox of functions used in the Simulator class to calculate key bioelectric
# properties.

def electroflux(cA,cB,Dc,d,zc,vBA,T,p,rho=1):
    """
    Electro-diffusion between two connected volumes. Note for cell work, 'b' is
    'inside', 'a' is outside, with a positive flux moving from a to b. The
    voltage is defined as Vb - Va (Vba), which is equivalent to Vmem.

    This function defaults to regular diffusion if Vba == 0.0.

    This function takes numpy matrix values as input. All inputs must be
    matrices of the same shape.

    This is the Goldman Flux/Current Equation (not to be confused with the
    Goldman Equation). Note: the Nernst-Planck equation has been trialed in
    place of this, and it does not reproduce proper reversal potentials.

    Parameters
    ----------
    cA          concentration in region A [mol/m3] (out)
    cB          concentration in region B [mol/m3] (in)
    Dc          Diffusion constant of c  [m2/s]
    d           Distance between region A and region B [m]
    zc          valence of ionic species c
    vBA         voltage difference between region B (in) and A (out) = Vmem
    p           an instance of the Parameters class

    Returns
    --------
    flux        Chemical flux magnitude between region A and B [mol/s]

    """

    # Reasonably small real number, preventing divide by zero errors.
    # Note that "betse.util.type.numeric.floats.FLOAT_MIN", the
    # smallest possible real number, is too small for this use case.
    FLOAT_NONCE = 1.0e-25

    vBA += FLOAT_NONCE

    zc += FLOAT_NONCE

    alpha = (zc*vBA*p.F)/(p.R*T)

    exp_alpha = np.exp(-alpha)

    deno = -np.expm1(-alpha)   # calculate the denominator for the electrodiffusion equation,..

    # calculate the flux for those elements:
    flux = -((Dc*alpha)/d)*((cB -cA*exp_alpha)/deno)*rho

    # flux = flux*rho

    return flux

def pumpNaKATP(cNai,cNao,cKi,cKo,Vm,T,p,block, met = None):

    """
    Parameters
    ----------
    cNai            Concentration of Na+ inside the cell
    cNao            Concentration of Na+ outside the cell
    cKi             Concentration of K+ inside the cell
    cKo             Concentration of K+ outside the cell
    Vm              Voltage across cell membrane [V]
    p               An instance of Parameters object

    met             A "metabolism" vector containing concentrations of ATP, ADP and Pi


    Returns
    -------
    f_Na            Na+ flux (into cell +)
    f_K             K+ flux (into cell +)
    """

    deltaGATP_o = p.deltaGATP  # standard free energy of ATP hydrolysis reaction in J/(mol K)

    cATP = p.cATP
    cADP = p.cADP
    cPi  = p.cPi

    # calculate the reaction coefficient Q:
    Qnumo = (cADP*1e-3)*(cPi*1e-3)*((cNao*1e-3)**3)*((cKi*1e-3)**2)
    Qdenomo = (cATP*1e-3)*((cNai*1e-3)**3)*((cKo*1e-3)** 2)

    # ensure no chance of dividing by zero:
    inds_Z = (Qdenomo == 0.0).nonzero()
    Qdenomo[inds_Z] = 1.0e-15

    Q = Qnumo / Qdenomo


    # calculate the equilibrium constant for the pump reaction:
    Keq = np.exp(-(deltaGATP_o / (p.R * T) - ((p.F * Vm) / (p.R * T))))

    # calculate the enzyme coefficient:
    numo_E = ((cNai/p.KmNK_Na)**3) * ((cKo/p.KmNK_K)**2) * (cATP/p.KmNK_ATP)
    denomo_E = (1 + (cNai/p.KmNK_Na)**3)*(1+(cKo/p.KmNK_K)**2)*(1+(cATP/p.KmNK_ATP))

    fwd_co = numo_E/denomo_E

    f_Na = -3*block*p.alpha_NaK*fwd_co*(1 - (Q/Keq))  # flux as [mol/m2s]   scaled to concentrations Na in and K out

    f_K = -(2/3)*f_Na          # flux as [mol/m2s]

    return f_Na, f_K, -f_Na  # FIXME get rid of this return of extra -f_Na!!

def pumpCaATP(cCai,cCao,Vm,T,p, block, met = None):

    """
    Parameters
    ----------
    cCai            Concentration of Ca2+ inside the cell
    cCao            Concentration of Ca2+ outside the cell
    voli            Volume of the cell [m3]
    volo            Volume outside the cell [m3]
    Vm              Voltage across cell membrane [V]
    p               An instance of Parameters object


    Returns
    -------
    cCai2           Updated Ca2+ inside cell
    cCao2           Updated Ca2+ outside cell
    f_Ca            Ca2+ flux (into cell +)
    """


    deltaGATP_o = p.deltaGATP

    no_negs(cCai)
    no_negs(cCao)

    cATP = p.cATP
    cADP = p.cADP
    cPi  = p.cPi
    #

    # calculate the reaction coefficient Q:
    Qnumo = cADP * cPi * cCao
    Qdenomo = cATP * cCai

    # ensure no chance of dividing by zero:
    inds_Z = (Qdenomo == 0.0).nonzero()
    Qdenomo[inds_Z] = 1.0e-16

    Q = Qnumo / Qdenomo

    # calculate the equilibrium constant for the pump reaction:
    Keq = np.exp(-(deltaGATP_o / (p.R * T) - 2*((p.F * Vm) / (p.R * T))))

    # calculate the enzyme coefficient for forward reaction:
    numo_E = (cCai/p.KmCa_Ca) * (cATP/p.KmCa_ATP)
    denomo_E = (1 + (cCai/p.KmCa_Ca)) * (1+ (cATP/p.KmCa_ATP))

    frwd = numo_E/denomo_E

    # calculate the enzyme coefficient for backward reaction:
    numo_Eb = (cCao/p.KmCa_Ca)
    denomo_Eb = (1 + (cCao/p.KmCa_Ca))

    bkwrd = numo_Eb/denomo_Eb

    f_Ca = -p.alpha_Ca*frwd*(1 - (Q/Keq))  # flux as [mol/m2s]

    return f_Ca

def pumpCaER(cCai,cCao,Vm,T,p):
    """
    Pumps calcium out of the cell and into the endoplasmic reticulum.
    Vm is the voltage across the endoplasmic reticulum membrane.

    """

    deltaGATP_o = p.deltaGATP

    cATP = p.cATP
    cADP = p.cADP
    cPi = p.cPi

    # calculate the reaction coefficient Q:
    Qnumo = cADP * cPi * cCai
    Qdenomo = cATP * cCao

    # ensure no chance of dividing by zero:
    inds_Z = (Qdenomo == 0.0).nonzero()
    Qdenomo[inds_Z] = 1.0e-16

    Q = Qnumo / Qdenomo

    # calculate the equilibrium constant for the pump reaction:
    Keq = np.exp(-deltaGATP_o / (p.R * T) - 2 * ((p.F * Vm) / (p.R * T)))


    # calculate the enzyme coefficient for forward reaction:
    numo_E = (cCao / p.KmCa_Ca) * (cATP / p.KmCa_ATP)
    denomo_E = (1 + (cCao / p.KmCa_Ca)) * (1 + (cATP / p.KmCa_ATP))

    frwd = numo_E / denomo_E

    # calculate the enzyme coefficient for backward reaction:
    numo_Eb = (cCai / p.KmCa_Ca)
    denomo_Eb = (1 + (cCai / p.KmCa_Ca))

    bkwrd = numo_Eb / denomo_Eb

    f_Ca = p.serca_max * frwd * (1 - (Q / Keq))  # flux as [mol/m2s]

    return f_Ca

def get_volt(q,sa,p):

    """
    Calculates the voltage for a net charge on a capacitor.

    Parameters
    ----------
    q           Net electrical charge [C]
    sa          Surface area [m2]

    Returns
    -------
    V               Voltage on the capacitive space holding charge

    """

    cap = sa*p.cm
    V = (1/cap)*q

    return V

def get_charge(concentrations,zs,vol,p):
    """
    Calculates the total charge in a space given a set of concentrations
    and their ionic charges, along with the space volume.

    Parameters
    ---------------
    concentrations       An array of arrays listing concentrations of different ions in multiple spaces [mol/m2].
    zs                   An array of ion charge states
    vol                  Volume of the spaces [m3]
    p                    Instance of Parameters module object

    Returns
    ----------------
    netcharge            The unbalanced charge in the space or spaces [C]

    """

    netcharge = np.sum(p.F*vol*concentrations*zs,axis=0)

    return netcharge

def get_charge_density(concentrations,zs,p):

    """
    Calculates the charge density given ion concentrations in an array of spaces.

    Parameters
    --------------
    concentrations:  an array of array of concentration of ions in spaces [mol/m3]
    zs:              valence of each ion
    p:               Parameters object instance

    Returns
    -------------
    netcharge     the net charge density in spaces C/m3
    """

    netcharge = np.sum(p.F*zs*concentrations, axis=0)

    return netcharge

def get_molarity(concentrations,p):

    q = 0

    for conc in concentrations:
        q = q+ conc

    netmolarity = q

    return netmolarity

def cell_ave(cells,vm_at_mem):

    """
    Averages Vmem over membrane domains to return a mean value for each cell

    Parameters
    ----------
    cells               An instance of the Cells module cells object
    vm_at_mem           Vmem at individual membrane domains


    Returns
    --------
    v_cell              Cell Vm averaged over the whole cell

    """

    v_cell = []

    for i in cells.cell_i:
        cellinds = (cells.mem_to_cells == i).nonzero()
        v_cell_array = vm_at_mem[cellinds]
        v_cell_ave = np.mean(v_cell_array)
        v_cell.append(v_cell_ave)

    v_cell = np.asarray(v_cell)

    return v_cell

#FIXME: For efficiency, all calls to this function should be replaced by calls
#to the dramatically faster betse.lib.numpy.nptest.die_if_nan() function. Hola!
def check_v(vm):
    """
    Does a quick check on Vmem values
    and displays error warning or exception if the value
    indicates the simulation is unstable.
    """

    isnans = np.isnan(vm)

    if isnans.any():  # if there's anything in the isubzeros matrix...
        raise BetseSimUnstableException(
            "Your simulation has become unstable. Please try a smaller time step,"
            "reduce gap junction radius, and/or reduce pump rate coefficients.")

def vertData(data, cells, p):
    """
    Interpolate data from midpoints to verts
    and sample it over a uniform grid.
    Produces data suitable for 2D mesh and
    streamline plots.

    Parameters
    -----------
    data          A numpy vector of data points on cell membrane mids
    cells         An instance of the Cells object
    p             An instance of the Parameters object

    Returns
    ----------
    dat_grid      THe data sampled on a uniform grid
    """

    # interpolate vmem defined on mem mids to cell vertices:
    verts_data = np.dot(data,cells.matrixMap2Verts)

    # amalgamate both mem mids and verts data into one stack:
    plot_data = np.hstack((data,verts_data))

    # interpolate the stack to the plotting grid:
    dat_grid = interp.griddata((cells.plot_xy[:,0],cells.plot_xy[:,1]),plot_data,(cells.Xgrid,cells.Ygrid),
                               method=p.interp_type,
                               fill_value=0)

    # # smooth out the data a bit:
    dat_grid = gaussian_filter(dat_grid,1)

    # get rid of values that bleed into the environment:
    # dat_grid = np.multiply(dat_grid,cells.maskM)

    dat_grid = ma.masked_array(dat_grid, np.logical_not(cells.maskM))

    return dat_grid

def nernst_planck_flux(c, gcx, gcy, gvx, gvy,ux,uy,D,z,T,p, mu = 0.0):
    """
     Calculate the flux component of the Nernst-Planck equation

     Parameters
     ------------

    c:     concentration
    gcx:   concentration gradient, x component
    gcy:   concentration gradient, y component
    gvx:   voltage gradient, x component
    gvy:   voltage gradient, y component
    ux:    fluid velocity, x component
    uy:    fluid velocity, y component
    D:     diffusion constant, D
    z:     ion charge
    T:     temperature
    p:     parameters object

    Returns
    --------
    fx, fx        mass flux in x and y directions
    """

    alpha = (D*z*p.q)/(p.kb*T)
    fx =  -D*gcx - alpha*gvx*c + ux*c - mu*c*gvx
    fy =  -D*gcy - alpha*gvy*c + uy*c - mu*c*gvy

    return fx, fy

def nernst_planck_vector(c, gc, gv,u,D,z,T,p):
    """
     Calculate the flux component of the Nernst-Planck equation
     along a directional gradient (e.g. gap junction)

     Parameters
     ------------

    c:     concentration
    gc:   concentration gradient
    gv:   voltage gradient
    u:    fluid velocity
    D:     diffusion constant, D
    z:     ion charge
    T:     temperature
    p:     parameters object

    Returns
    --------
    fx, fx        mass flux in x and y directions
    """

    alpha = (D*z*p.q)/(p.kb*T)
    f =  -D*gc - alpha*gv*c + u*c

    return f

def np_flux_special(cx,cy,gcx,gcy,gvx,gvy,ux,uy,Dx,Dy,z,T,p):
    """
     Calculate the flux component of the Nernst-Planck equation on a MACs grid

     Parameters
     ------------

    c:     concentration
    gcx:   concentration gradient, x component
    gcy:   concentration gradient, y component
    gvx:   voltage gradient, x component
    gvy:   voltage gradient, y component
    ux:    fluid velocity, x component
    uy:    fluid velocity, y component
    D:     diffusion constant, D
    z:     ion charge
    T:     temperature
    p:     parameters object

    Returns
    --------
    fx, fx        mass flux in x and y directions
    """

    alphax = (Dx*z*p.q)/(p.kb*T)
    alphay = (Dy*z*p.q)/(p.kb*T)

    fx =  -Dx*gcx - alphax*gvx*cx + cx*ux

    fy =  -Dy*gcy - alphay*gvy*cy + cy*uy

    return fx, fy

def no_negs(data):
    """
    This function screens an (concentration) array to
    ensure there are no NaNs and no negative values,
    crashing with an instability message if it finds any.
    """

    #FIXME: For efficiency, this NaN check should be replaced by a call to the
    #dramatically faster betse.lib.numpy.nptest.die_if_nan() function. O arise!

    # Ensure no NaNs.
    inds_nan = (np.isnan(data)).nonzero()

    # Ensure that data has no negative values.
    inds_neg = (data < 0.0).nonzero()

    #FIXME: This seems to contradict the documentation. It also seems a bit
    #unsafe. Shouldn't this raise an exception rather than silently replace all
    #negative values with 0.0? Or maybe this is O.K.? Cloudy marshmallows!
    if len(inds_neg[0]) > 0:
        data[inds_neg] = 0.0 # add in a small bit to protect from crashing

    if len(inds_nan[0]) > 0:

        raise BetseSimUnstableException(
            "Your simulation has become unstable. Please try a smaller time step,"
            "reduce gap junction radius, and/or reduce rate coefficients.")

    return data

def bicarbonate_buffer(cCO2, cHCO3):
    """
    This most amazing buffer handles influx of H+,
    HCO3-, H2CO3 (from dissolved carbon dioxide) to
    handle pH in real time.

    Uses the bicarbonate dissacociation reaction:

    H2CO3 ----> HCO3 + H

    Where all dissolved carbon dioxide is assumed
    converted to carbonic acid via carbonic anhydrase enzyme.

    """

    pH = 6.1 + np.log10(cHCO3/cCO2)

    cH = 10**(-pH)*1e3

    return cH, pH

def ghk_calculator(sim, cells, p):
    """
    Uses simulation parameters in the Goldman (GHK) equation
    to calculate an alternative Vmem for validation purposes.

    """


    # FIXME the Goldman calculator must be altered to account for network pumps and channels!!
    # begin by initializing all summation arrays for the cell network:
    sum_PmAnion_out = []
    sum_PmAnion_in = []
    sum_PmCation_out = []
    sum_PmCation_in = []

    for i, z in enumerate(sim.zs):

        # tag as anion or cation
        ion_type = np.sign(z)

        # average values from membranes or environment to cell centres:
        Dm = np.dot(cells.M_sum_mems, sim.Dm_cells[i]) / cells.num_mems
        conc_cells = sim.cc_cells[i]

        if p.is_ecm is True:
            # average entities from membranes to the cell centres:
            conc_env = np.dot(cells.M_sum_mems, sim.cc_env[i][cells.map_mem2ecm]) / cells.num_mems

        else:

            conc_env = np.dot(cells.M_sum_mems, sim.cc_env[i]) / cells.num_mems

        if ion_type == -1:

            sum_PmAnion_in.append(Dm * conc_cells * (1 / p.tm))
            sum_PmAnion_out.append(Dm * conc_env * (1 / p.tm))


        if ion_type == 1:

            sum_PmCation_in.append(Dm * conc_cells * (1 / p.tm))
            sum_PmCation_out.append( Dm * conc_env * (1 / p.tm))

    if p.molecules_enabled:

        for name in sim.molecules.core.channels:
            obj = sim.molecules.core.channels[name]

            for ii, relP in zip(obj.channel_core.ions, obj.channel_core.rel_perm):

                ion_i = sim.get_ion(ii)
                zi = sim.zs[ion_i]
                conc_cells = sim.cc_cells[ion_i]

                # tag as anion or cation
                ion_type = np.sign(zi)

                if p.is_ecm is True:
                    # average entities from membranes to the cell centres:
                    conc_env = np.dot(cells.M_sum_mems, sim.cc_env[ion_i][cells.map_mem2ecm]) / cells.num_mems

                else:

                    conc_env = np.dot(cells.M_sum_mems, sim.cc_env[ion_i]) / cells.num_mems

                if obj.channel_core.DChan is not None:
                    Dmo = obj.channel_core.DChan*relP
                    Dm = np.dot(cells.M_sum_mems, Dmo) / cells.num_mems

                else:
                    Dm = 0.0

                if ion_type == -1:
                    sum_PmAnion_in.append(Dm * conc_cells * (1 / p.tm))
                    sum_PmAnion_out.append(Dm * conc_env * (1 / p.tm))

                if ion_type == 1:
                    sum_PmCation_in.append( Dm * conc_cells * (1 / p.tm))
                    sum_PmCation_out.append(Dm * conc_env * (1 / p.tm))

    if p.grn_enabled:

        for name in sim.grn.core.channels:
            obj = sim.grn.core.channels[name]

            for ii, relP in zip(obj.channel_core.ions, obj.channel_core.rel_perm):

                ion_i = sim.get_ion(ii)
                zi = sim.zs[ion_i]
                conc_cells = sim.cc_cells[ion_i]

                # tag as anion or cation
                ion_type = np.sign(zi)

                if p.is_ecm is True:
                    # average entities from membranes to the cell centres:
                    conc_env = np.dot(cells.M_sum_mems,
                                      sim.cc_env[ion_i][cells.map_mem2ecm]) / cells.num_mems

                else:

                    conc_env = np.dot(cells.M_sum_mems, sim.cc_env[ion_i]) / cells.num_mems

                if obj.channel_core.DChan is not None:
                    Dmo = obj.channel_core.DChan * relP
                    Dm = np.dot(cells.M_sum_mems, Dmo) / cells.num_mems

                else:
                    Dm = 0.0

                if ion_type == -1:
                    sum_PmAnion_in.append(Dm * conc_cells * (1 / p.tm))
                    sum_PmAnion_out.append(Dm * conc_env * (1 / p.tm))

                if ion_type == 1:
                    sum_PmCation_in.append(Dm * conc_cells * (1 / p.tm))
                    sum_PmCation_out.append(Dm * conc_env * (1 / p.tm))


    # NaKrate = (np.dot(cells.M_sum_mems, sim.rate_NaKATP)/cells.num_mems)

    # sum together contributions for Na and K flux across the membrane:
    # NaKflux = NaKrate - (2/3)*NaKrate

    sum_PmAnion_in_i = np.sum(sum_PmAnion_in, axis = 0)
    sum_PmAnion_out_i = np.sum(sum_PmAnion_out, axis=0)
    sum_PmCation_in_i = np.sum(sum_PmCation_in, axis=0)
    sum_PmCation_out_i = np.sum(sum_PmCation_out, axis=0)

    sim.vm_GHK = ((p.R * sim.T) / p.F) * np.log(
        (sum_PmCation_out_i + sum_PmAnion_in_i) / (sum_PmCation_in_i + sum_PmAnion_out_i))

def molecule_pump(sim, cX_cell_o, cX_env_o, cells, p, Df=1e-9, z=0, pump_into_cell =False, alpha_max=1.0e-8, Km_X=1.0,
                 Km_ATP=1.0, met = None, n=1, ignoreECM = True, rho = 1.0):


    """
    Defines a generic active transport pump that can be used to move
    a general molecule (such as serotonin or glutamate)
    into or out of the cell by active transport.

    Works on the basic premise of enzymatic pumps defined elsewhere:

    pump_out is True:

    cX_cell + cATP  -------> cX_env + cADP + cPi

    pump_out is False:

    cX_env + cATP  <------- cX_cell + cADP + cPi

    Parameters
    -------------
    cX_cell_o           Concentration of X in the cell         [mol/m3]
    cX_env_o            Concentration of X in the environment  [mol/m3]
    cells               Instance of cells
    p                   Instance of parameters
    z                   Charge of X
    pump_out            Is pumping out of cell (pump_out = True) or into cell (pump_out = False)?
    alpha_max           Maximum rate constant of pump reaction [mol/s]
    Km_X                Michaelis-Mentin 1/2 saturation value for X [mol/m3]
    Km_ATP              Michaelis-Mentin 1/2 saturation value for ATP [mol/m3]

    Returns
    ------------
    cX_cell_1     Updated concentration of X in cells
    cX_env_1      Updated concentration of X in environment
    f_X           Flux of X (into the cell +)

    """

    deltaGATP_o = p.deltaGATP  # standard free energy of ATP hydrolysis reaction in J/(mol K)

    if met is None:

        # if metabolism vector not supplied, use singular defaults for concentrations
        cATP = p.cATP
        cADP = p.cADP
        cPi  = p.cPi

    else:

        cATP = met['cATP']  # concentration of ATP in mmol/L
        cADP = met['cADP']  # concentration of ADP in mmol/L
        cPi = met['cPi']  # concentration of Pi in mmol/L

    if p.is_ecm is True:

        cX_env = cX_env_o[cells.map_mem2ecm]

        cX_cell = cX_cell_o[cells.mem_to_cells]

    else:
        cX_env = cX_env_o[:]
        cX_cell = cX_cell_o[cells.mem_to_cells]

    if pump_into_cell is False:

        # active pumping of molecule from cell and into environment:
        # calculate the reaction coefficient Q:
        Qnumo = cADP * cPi * (cX_env**n)
        Qdenomo = cATP * (cX_cell**n)

        # ensure no chance of dividing by zero:
        inds_Z = (Qdenomo == 0.0).nonzero()
        Qdenomo[inds_Z] = 1.0e-10

        Q = Qnumo / Qdenomo

        # calculate the equilibrium constant for the pump reaction:
        Keq = np.exp(-deltaGATP_o / (p.R * sim.T) + ((n*z * p.F * sim.vm) / (p.R * sim.T)))

        # calculate the reaction rate coefficient
        alpha = alpha_max * (1 - (Q / Keq))

        # calculate the enzyme coefficient:
        numo_E = ((cX_cell / Km_X)**n) * (cATP / Km_ATP)
        denomo_E = (1 + (cX_cell / Km_X)**n) * (1 + (cATP / Km_ATP))

        f_X = -rho*alpha * (numo_E / denomo_E)  # flux as [mol/m2s]   scaled to concentrations Na in and K out

    else:

        # active pumping of molecule from environment and into cell:
        # calculate the reaction coefficient Q:
        Qnumo = cADP * cPi * cX_cell
        Qdenomo = cATP * cX_env

        # ensure no chance of dividing by zero:
        inds_Z = (Qdenomo == 0.0).nonzero()
        Qdenomo[inds_Z] = 1.0e-10

        Q = Qnumo / Qdenomo

        # calculate the equilibrium constant for the pump reaction:
        Keq = np.exp(-deltaGATP_o / (p.R * sim.T) - ((z * p.F * sim.vm) / (p.R * sim.T)))

        # calculate the reaction rate coefficient
        alpha = alpha_max * (1 - (Q / Keq))

        # calculate the enzyme coefficient:
        numo_E = (cX_env / Km_X) * (cATP / Km_ATP)
        denomo_E = (1 + (cX_env / Km_X)) * (1 + (cATP / Km_ATP))

        f_X = rho* alpha * (numo_E / denomo_E)  # flux as [mol/m2s]   scaled to concentrations Na in and K out

    if p.cluster_open is False:
        f_X[cells.bflags_mems] = 0

    cmems = cX_cell_o[cells.mem_to_cells]

    # update cell and environmental concentrations
    cX_cell_1, _, cX_env_1 = update_Co(sim, cX_cell_o, cmems, cX_env_o, f_X, cells, p, ignoreECM = ignoreECM)


    if p.is_ecm is False:
        cX_env_1_temp = cX_env_1.mean()
        cX_env_1[:] = cX_env_1_temp

    return cX_cell_1, cX_env_1, f_X

def molecule_transporter(sim, cX_cell_o, cX_env_o, cells, p, Df=1e-9, z=0, pump_into_cell=False, alpha_max=1.0e-8,
        Km_X=1.0, Keq=1.0, n = 1.0, ignoreECM = True, rho = 1.0):


    """
    Defines a generic facillitated transporter that can be used to move
    a general molecule (such as glucose).

    ATP is not used for the transporter

    Works on the basic premise of enzymatic pumps defined elsewhere:

    pump_out is True:

    cX_cell  -------> cX_env

    pump_out is False:

    cX_env   <------- cX_cell

    Parameters
    -------------
    cX_cell_o           Concentration of X in the cell         [mol/m3]
    cX_env_o            Concentration of X in the environment  [mol/m3]
    cells               Instance of cells
    p                   Instance of parameters
    z                   Charge of X
    pump_out            Is pumping out of cell (pump_out = True) or into cell (pump_out = False)?
    alpha_max           Maximum rate constant of pump reaction [mol/s]
    Km_X                Michaelis-Mentin 1/2 saturation value for X [mol/m3]

    Returns
    ------------
    cX_cell_1     Updated concentration of X in cells
    cX_env_1      Updated concentration of X in environment
    f_X           Flux of X (into the cell +)

    """

    if p.is_ecm is True:

        cX_env = cX_env_o[cells.map_mem2ecm]

        cX_cell = cX_cell_o[cells.mem_to_cells]

    else:
        cX_env = cX_env_o
        cX_cell = cX_cell_o[cells.mem_to_cells]

    if pump_into_cell is False:

        # active pumping of molecule from cell and into environment:
        # calculate the reaction coefficient Q:
        Qnumo = cX_env
        Qdenomo = cX_cell

        # ensure no chance of dividing by zero:
        inds_Z = (Qdenomo == 0.0).nonzero()
        Qdenomo[inds_Z] = 1.0e-15

        Q = Qnumo / Qdenomo

        # modify equilibrium constant by membrane voltage if ion is charged:
        Keq = Keq*np.exp((z * p.F * sim.vm) / (p.R * sim.T))

        # calculate the reaction rate coefficient
        alpha = alpha_max * (1 - (Q / Keq))

        # calculate the enzyme coefficient:
        numo_E = (cX_cell / Km_X)
        denomo_E = (1 + (cX_cell / Km_X))

        f_X = -rho*alpha * (numo_E / denomo_E)  # flux as [mol/m2s]   scaled to concentrations Na in and K out


    else:

        # active pumping of molecule from environment and into cell:
        # calculate the reaction coefficient Q:
        Qnumo = cX_cell
        Qdenomo = cX_env

        # ensure no chance of dividing by zero:
        inds_Z = (Qdenomo == 0.0).nonzero()
        Qdenomo[inds_Z] = 1.0e-15

        Q = Qnumo / Qdenomo

        # modify equilibrium constant by membrane voltage if ion is charged:
        Keq = Keq*np.exp(-(z * p.F * sim.vm) / (p.R * sim.T))

        # calculate the reaction rate coefficient
        alpha = alpha_max * (1 - (Q / Keq))

        # calculate the enzyme coefficient:
        numo_E = (cX_env / Km_X)
        denomo_E = (1 + (cX_env / Km_X))

        f_X = rho*alpha * (numo_E / denomo_E)  # flux as [mol/m2s]   scaled to concentrations Na in and K out

    if p.cluster_open is False:
        f_X[cells.bflags_mems] = 0

    cmems = cX_cell_o[cells.mem_to_cells]

    # update cell and environmental concentrations
    cX_cell_1, _, cX_env_1 = update_Co(sim, cX_cell_o, cmems, cX_env_o, f_X, cells, p, ignoreECM= ignoreECM)

    # next electrodiffuse concentrations around the cell interior:
    # cX_cell_1 = update_intra(sim, cells, cX_cell_1, Df, z, p)

    # ensure that there are no negative values
    # cX_cell_1 = no_negs(cX_cell_1)
    # cX_env_1 = no_negs(cX_env_1)


    if p.is_ecm is False:
        cX_env_1_temp = cX_env_1.mean()
        cX_env_1[:] = cX_env_1_temp

    return cX_cell_1, cX_env_1, f_X

def molecule_mover(sim, cX_env_o, cX_cells, cells, p, z=0, Dm=1.0e-18, Do=1.0e-9, Dgj=1.0e-12, Ftj = 1.0, c_bound=0.0,
                   ignoreECM = True, smoothECM = False, ignoreTJ = False, ignoreGJ = False, rho = 1, cmems = None,
                   time_dilation_factor = 1.0, update_intra = False, name = "Unknown", transmem = False,
                   mu_mem = 0.0, umt = 0.0):

    """
    Transports a generic molecule across the membrane,
    through gap junctions, and if p.is_ecm is true,
    through extracellular spaces and the environment.

    Parameters
    -----------
    cX_cell_o           Concentration of molecule in the cytosol [mol/m3]
    cX_env_o            Concentration of molecule in the environment [mol/m3]
    cells               Instance of Cells
    p                   Instance of Parameters
    z                   Charge state of molecule
    Dm                  Membrane diffusion constant [m2/s]
    Do                  Free diffusion constant [m2/s]
    Ftj                 Factor influencing relative diffusion of substance across tight junction barrier
    c_bound             Concentration of molecule at global bounds (required for is_ecm True only)

    Returns
    -----------
    cX_cell_1         Updated concentration of molecule in the cell
    cX_env_1          Updated concentration of molecule in the environment

    """


    if p.is_ecm is True:

        cX_env = cX_env_o[cells.map_mem2ecm]

    else:
        cX_env = cX_env_o[:]

    if cmems is None:

        cX_mems = cX_cells[cells.mem_to_cells]

    elif len(cmems) == sim.mdl:
        cX_mems = cmems

    Dm_vect = np.ones(len(cX_mems))*Dm

    if Dm != 0.0:  # if there is some finite membrane diffusivity, exchange between cells and env space:

        IdM = np.ones(sim.mdl)

        f_X_ED = electroflux(cX_env, cX_mems, Dm_vect, p.tm*IdM, z*IdM, sim.vm, sim.T, p, rho = rho)

        if p.cluster_open is False:
            f_X_ED[cells.bflags_mems] = 0

    else:

        f_X_ED = np.zeros(sim.mdl)

    # update concentrations due to electrodiffusion:
    cX_cells, cX_mems, cX_env_o = update_Co(sim, cX_cells, cX_mems, cX_env_o, f_X_ED, cells, p,
                                            ignoreECM=ignoreECM, update_at_mems = update_intra)

    # ------------------------------------------------------------
    if ignoreGJ is False:

        # Update gap junctions using the GHK-flux equation
        fgj_X = electroflux(cX_mems[cells.mem_i],
                       cX_mems[cells.nn_i],
                       Dgj*sim.gj_block*sim.gjopen,
                       cells.gj_len*np.ones(sim.mdl),
                       z*np.ones(sim.mdl),
                       sim.vgj,
                       p.T,
                       p,
                       rho=1
                       )

        # enforce zero flux at outer boundary:
        fgj_X[cells.bflags_mems] = 0.0

        delta_cco = np.dot(cells.M_sum_mems, -fgj_X * cells.mem_sa) / cells.cell_vol

        # Calculate the final concentration change (the acceleration effectively speeds up time):
        if update_intra is False: # do the GJ transfer assuming instant mixing in the cell:
            # divergence calculation for individual cells (finite volume expression)
            cX_cells = cX_cells + p.dt*delta_cco*time_dilation_factor
            cX_mems = cX_cells[cells.mem_to_cells]

        else: # only update the membranes; calculate cell centre as the average:
            cX_cells = cX_cells + p.dt * delta_cco * time_dilation_factor
            # the 3/4 is the volume of the pie-shaped wedge half-triangle alloted to specific membrane volume:
            cX_mems += -fgj_X*(cells.mem_sa/((3/4)*cells.mem_vol))*time_dilation_factor*p.dt

    else:
        fgj_X = np.zeros(sim.mdl)

    #----Motor protein transport-----------------------------------------------------------------------------
    if update_intra is False and ignoreGJ and transmem is True:

        nx = cells.mem_vects_flat[:, 2]
        ny = cells.mem_vects_flat[:, 3]

        # concentration gradient at cell centres:
        gcc, gccmx, gccmy = cells.gradient(cX_cells)

        # mean value of concentration between two cells:
        mcc = (cX_mems[cells.nn_i] + cX_mems[cells.mem_i])/2
        # mcc = cX_cells[cells.mem_to_cells]

        # mean value of u-field between cells, treating motor protein transport like convection field:
        # umtx = (sim.mtubes.mtubes_x[cells.nn_i] + sim.mtubes.mtubes_x[cells.mem_i])/2
        # umty = (sim.mtubes.mtubes_y[cells.nn_i] + sim.mtubes.mtubes_y[cells.mem_i])/2
        umtx = 0.0
        umty = 0.0

        # umtx = sim.mtubes.uxmt[cells.mem_to_cells]
        # umty = sim.mtubes.uymt[cells.mem_to_cells]

        emtx = (sim.E_cell_x[cells.mem_to_cells][cells.nn_i] + sim.E_cell_x[cells.mem_to_cells][cells.nn_i])/2
        emty = (sim.E_cell_y[cells.mem_to_cells][cells.nn_i] + sim.E_cell_y[cells.mem_to_cells][cells.nn_i])/2

        # emtx = sim.E_cell_x[cells.mem_to_cells]
        # emty = sim.E_cell_y[cells.mem_to_cells]

        alpha = (Do*p.q*z)/(p.kb*sim.T)

        # bgrad_x = ((umtx[cells.bflags_mems]*umt + (mu_mem + alpha)*emtx[cells.bflags_mems])*mcc[cells.bflags_mems])/Do
        # bgrad_y = ((umty[cells.bflags_mems]*umt + (mu_mem + alpha)*emty[cells.bflags_mems])*mcc[cells.bflags_mems])/Do
        #
        # gccmx[cells.bflags_mems] = bgrad_x
        # gccmy[cells.bflags_mems] = bgrad_y

        flux_mtx = -Do*gccmx + umtx*mcc*umt + (mu_mem + alpha)*emtx*mcc
        flux_mty = -Do*gccmy + umty*mcc*umt + (mu_mem + alpha)*emty*mcc

        # flux_mtx[cells.bflags_mems] = -umtx[cells.bflags_mems]*mcc[cells.bflags_mems]*umt
        # flux_mty[cells.bflags_mems] = -umty[cells.bflags_mems]*mcc[cells.bflags_mems]*umt

        flux_mtn = flux_mtx*nx  + flux_mty*ny
        # flux_mtn = -Do*gcc + sim.mtubes.umtn*mcc*umt

        flux_mtn[cells.bflags_mems] = 0.0

        div_ccmt = -np.dot(cells.M_sum_mems, flux_mtn*cells.mem_sa)/cells.cell_vol

        # update cell concentration:
        cX_cells += div_ccmt*p.dt*time_dilation_factor

        cX_mems = cX_cells[cells.mem_to_cells]*1



    # Transport through environment, if p.is_ecm is True-----------------------------------------------------
    if p.is_ecm is True:

        env_check = len((cX_env_o != 0.0).nonzero()[0])

        if  env_check != 0.0 or c_bound > 1.0e-15:

            cenv = cX_env_o
            cenv = cenv.reshape(cells.X.shape)

            cenv[:, 0] = c_bound
            cenv[:, -1] = c_bound
            cenv[0, :] = c_bound
            cenv[-1, :] = c_bound

            denv_multiplier = np.ones(len(cells.xypts))

            if ignoreTJ is False:
                # if tight junction barrier applies, create a mask that defines relative strength of barrier:
                denv_multiplier = denv_multiplier.reshape(cells.X.shape)*sim.D_env_weight

                # at the cluster boundary, further modify the env diffusion map by a relative TJ diffusion factor:
                denv_multiplier.ravel()[sim.TJ_targets] = denv_multiplier.ravel()[sim.TJ_targets]*Ftj

            else:
                denv_multiplier = denv_multiplier.reshape(cells.X.shape)

            gcx, gcy = fd.gradient(cenv, cells.delta)

            if p.fluid_flow is True:

                ux = sim.u_env_x.reshape(cells.X.shape)
                uy = sim.u_env_y.reshape(cells.X.shape)

            else:

                ux = 0.0
                uy = 0.0

            fx, fy = nernst_planck_flux(cenv, gcx, gcy, -sim.E_env_x, -sim.E_env_y, ux, uy,
                                            denv_multiplier*Do, z, sim.T, p, mu = mu_mem)

            div_fa = fd.divergence(-fx, -fy, cells.delta, cells.delta)

            fenvx = fx
            fenvy = fy

            cenv = cenv + div_fa * p.dt*time_dilation_factor

            if p.sharpness < 1.0:

                cenv = fd.integrator(cenv, sharp = p.sharpness)

            cX_env_o = cenv.ravel()

        else:

            fenvx = np.zeros(sim.edl)
            fenvy = np.zeros(sim.edl)

    else:
        cX_env_temp = cX_env_o.mean()
        cX_env_o = np.zeros(sim.mdl)
        cX_env_o[:] = cX_env_temp
        fenvx = 0
        fenvy = 0

    # check for sub-zero concentrations:
    indsZm = (cX_mems < 0.0).nonzero()[0]


    indsZc = (cX_cells < 0.0).nonzero()[0]

    if len(indsZc) > 0:
        raise BetseSimUnstableException(
            "Network concentration of " + name + " in cells below zero! Your simulation has"
                                                   " become unstable.")

    if len(indsZm) > 0:
        raise BetseSimUnstableException(
            "Network concentration of " + name + " on membrane below zero! Your simulation has"
                                                   " become unstable.")

    indsZe = (cX_env_o < 0.0).nonzero()[0]

    if len(indsZe) > 0:
        print(name, c_bound)
        raise BetseSimUnstableException(
            "Network concentration of " + name + " in environment below zero! Your simulation has"
                                                   " become unstable.")

    return cX_env_o, cX_cells, cX_mems, f_X_ED, fgj_X, fenvx, fenvy

def update_Co(sim, cX_cell, cX_mem, cX_env, flux, cells, p, ignoreECM = True, update_at_mems = False):
    """

    General updater for a concentration defined on
    cells and in environment.

    Parameters
    --------------
    cX_cell     Concentration inside cell defined on cell centres [mol/m3]
    cX_env      Concentration in extracellular spaces [mol/m3]
    flux        Flux of X, where into cell is + and component is normal to membranes  [mol/m2 s]
    cells       Instance of Cells
    p           Instance of Parameters

    Returns
    --------------
    cX_cell     Updated concentration in cell [mol/m3]
    cX_env      Updated concentration in environment [mol/m3]

    """

    # take the divergence of the flux for each enclosed cell:
    delta_cells = np.dot(cells.M_sum_mems, flux * cells.mem_sa) / cells.cell_vol

    # update cell concentration of substance:
    if update_at_mems is False: # treat cell mem and centre values as equal
        cX_cell = cX_cell + delta_cells * p.dt
        cX_mem = cX_cell[cells.mem_to_cells]

    else: # update only the membranes, treat cell centre as the average:
        cX_mem += flux*(cells.mem_sa/((3/4)*cells.mem_vol))*p.dt # the 3/4 is the volume of the pi wedge half-triangle
        cX_cell = cX_cell + delta_cells * p.dt


    if p.is_ecm is True:

        # delta_env = np.dot(cells.M_divmap_mem2ecm, -flux)
        delta_env = div_env(-flux, cells,p)

        # update the environmental concentrations:
        cX_env = cX_env + delta_env * p.dt


    else:

        delta_env = -flux * (cells.mem_sa / p.vol_env)

        cX_env_o = cX_env + delta_env * p.dt

        # assume auto-mixing of environmental concentrations:
        cX_env = cX_env_o.mean()

    return cX_cell, cX_mem, cX_env

def div_env(flux, cells, p):

    """
    Calculates the divergence of the normal net flux component defined on a set of cell membranes
    with respect to the environmental grid spaces.

    :param flux:
    :param cells:
    :return:
    """

    if p.fast_update_ecm:

        flux_env = np.zeros(len(cells.xypts))
        flux_env[cells.map_mem2ecm] = flux

        delta_env = (flux_env * cells.memSa_per_envSquare) / (cells.ecm_vol)

    else:
        # Method # 2:
        delta_env = np.dot(cells.M_divmap_mem2ecm, flux)/(p.cell_height*cells.delta**2)

        # use the "integrator" function to conservatively distribute this exchange to nearest neighbours of the env grid:
        # delta_env = fd.integrator(delta_env.reshape(cells.X.shape), sharp = 0.5).ravel()

    return delta_env

def HH_Decomp(JJx, JJy, cells, bounds = None, sigma = 1.0):

    if bounds is not None:

        Lb = bounds['L']
        Rb = bounds['R']
        Tb = bounds['T']
        Bb = bounds['B']

    else:
        Lb = 0.0
        Rb = 0.0
        Tb = 0.0
        Bb = 0.0

    # ----divergence-free component--------------------------------------

    Jxr = -JJy.reshape(cells.X.shape)
    Jyr = JJx.reshape(cells.X.shape)

    divJr = fd.divergence(Jxr, Jyr, cells.delta, cells.delta)

    divJr[:, 0] = 0.0
    divJr[:, -1] = 0.0
    divJr[0, :] = 0.0
    divJr[-1, :] = 0.0

    AA = np.dot(cells.lapENVinv, -divJr.ravel())

    gAx, gAy = fd.gradient(AA.reshape(cells.X.shape), cells.delta)

    Fx = -gAy
    Fy = gAx
    #
    # F = np.sqrt(Fx ** 2 + Fy ** 2)

    # ----curl free component------------------------------------------

    divJd = fd.divergence(JJx.reshape(cells.X.shape)/sigma, JJy.reshape(cells.X.shape)/sigma, cells.delta, cells.delta)

    # set boundary conditions for normal component
    # enforce applied voltage condition at the boundary:
    divJd[:, 0] = -Lb * (1 / cells.delta ** 2)
    divJd[:, -1] = -Rb * (1 / cells.delta ** 2)
    divJd[0, :] = -Bb * (1 / cells.delta ** 2)
    divJd[-1, :] = -Tb * (1 / cells.delta ** 2)


    BB = np.dot(cells.lapENVinv, divJd.ravel())

    Gx, Gy = fd.gradient(BB.reshape(cells.X.shape), cells.delta)

    # G = np.sqrt(Gx ** 2 + Gy ** 2)

    return AA, Fx, Fy, BB, Gx, Gy

def div_free(Fxo, Fyo, cells):

    """

    Uses a "hidden potential" method to
    calculate a divergence-free field

    :param Fxo:  x-component of input field
    :param Fyo:  y-component of input field
    :param cells:  cells instance
    :return: Fx, Fy divergence-free field components (finite divergence at bounds)

    """
    # calculate divergence of vector field:
    divF = fd.divergence(Fxo.reshape(cells.X.shape), Fyo.reshape(cells.X.shape), cells.delta, cells.delta)

    # value of the correcting potenial:
    Phi = np.dot(cells.lapENVinv, divF.ravel())

    gPhix, gPhiy = fd.gradient(Phi.reshape(cells.X.shape), cells.delta)

    Fx = Fxo.reshape(cells.X.shape) - gPhix
    Fy = Fyo.reshape(cells.X.shape) - gPhiy

    return Fx, Fy, Phi

def smooth_flux(Fxo, Fyo, cells):

    """
    Uses the Helmholtz-Hodge decomposition to smooth a vector field in a manner that does not change the physics.
    Fxo, Fyo must be arrays defined on the environmental grid

    """

    _, Frx, Fry, _, Fdx, Fdy = HH_Decomp(Fxo, Fyo, cells)

    # free current density at each membrane, "smoothed" using Helmholtz-Hodge decomposition and reconstruction:
    Fx = Frx + Fdx
    Fy = Fry + Fdy


    return Fx, Fy

def map_to_cells(Fx, Fy, cells, p, smoothing=1.0):
    """
    Takes a vector field defined on the environmental spaces and maps it to the cell grid.

    """
    if smoothing > 0.0:
        # smooth environmental electric field:
        F_env_x = gaussian_filter(Fx.reshape(cells.X.shape), smoothing).ravel()
        F_env_y = gaussian_filter(Fy.reshape(cells.X.shape), smoothing).ravel()

    else:

        F_env_x = Fx
        F_env_y = Fy

    # env electric field mapped to cells:
    Fx_atmem = F_env_x.ravel()[cells.map_mem2ecm]
    Fy_atmem = F_env_y.ravel()[cells.map_mem2ecm]

    # map env x to cell membrane component:
    Fmem = (Fx_atmem * cells.mem_vects_flat[:, 2] +
            Fy_atmem * cells.mem_vects_flat[:, 3])

    Fx_atcell = (np.dot(cells.M_sum_mems, Fx_atmem * cells.mem_sa) / cells.cell_sa)
    Fy_atcell = (np.dot(cells.M_sum_mems, Fy_atmem * cells.mem_sa) / cells.cell_sa)

    return Fx_atcell, Fy_atcell, Fmem

def get_conductivity(D, z, c, d, p):
    """
    Simple function returns conductivity (S/m)
    for diffusion constant D (m2/s), mean
    concentration c, and membrane thickness 'd'.

    """
    # (D * (z ** 2) * p.F * c) / (d * p.R * p.T)

    return (D * p.q *(z**2) * p.F * c) / (d * p.kb * p.T)

def rotate_field(gFxo, gFyo, angle):
    """
    Rotates a vector field gFxo, gFyo by a specified angle (in degrees)

    """
    # rotate the axis of the model:
    rotangle = (angle * np.pi) / 180.0

    gFx = gFxo * np.cos(rotangle) - gFyo * np.sin(rotangle)
    gFy = gFxo * np.sin(rotangle) + gFyo * np.cos(rotangle)

    return gFx, gFy






#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level support facilities for Yet Another Markup Language (YAML), the
file format encapsulating most input and output data for this application.
'''

#FIXME: Consider all or most of the related "yamlrepr" submodule back to
#"ruamel.yaml" -- especially the non-trivial Numpy-to-YAML-type conversions.

# ....................{ IMPORTS                           }....................
from betse.util.io import iofiles
from betse.util.io.error.errwarning import ignoring_warnings
from betse.util.io.log import logs
from betse.util.path import pathnames
from betse.util.type.contexts import noop_context
from betse.util.type.obj import objects
from betse.util.type.types import (
    type_check, MappingOrSequenceTypes, StrOrNoneTypes)
from ruamel import yaml as ruamel_yaml
from ruamel.yaml.error import MantissaNoDotYAML1_1Warning

# ....................{ GLOBALS                           }....................
YAML_FILETYPES = {'yaml', 'yml',}
'''
Set of all YAML-compliant filetypes.
'''

# ....................{ LOADERS                           }....................
@type_check
def load(
    # Mandatory parameters.
    filename: str,

    # Optional parameters.
    yaml_version: StrOrNoneTypes = None,
) -> MappingOrSequenceTypes:
    '''
    Load (i.e., open and read, deserialize) and return the contents of the
    YAML-formatted file with the passed path as either a dictionary or list
    via the active YAML implementation.

    Parameters
    ----------
    filename : str
        Absolute or relative filename of the YAML-formatted file to be loaded.
    yaml_version: StrOrNoneTypes
        Version of the YAML specification (e.g., 1.1, 1.2) to forcefully assume
        this file to be compliant with, overriding any version directive
        prefacing this file (e.g., ``%YAML 1.2``). Defaults to ``None``, in
        which case the version directive prefacing this file is deferred to.

    Returns
    ----------
    MappingOrSequenceTypes
        Dictionary or list corresponding to the contents of this file.
    '''

    # If this filename has no YAML-compliant filetype, log a warning.
    _warn_unless_filetype_yaml(filename)

    # With this YAML file opened for character-oriented reading...
    with iofiles.reading_chars(filename) as yaml_file:
        # Safe roundtripping YAML parser.
        ruamel_parser = _make_ruamel_parser()

        # Context manager with which to load this file from this parser,
        # defaulting to a noop context manager.
        context_manager = noop_context()

        # If assuming this file to comply with a specific version of the
        # YAML specification...
        if yaml_version is not None:
            # Inform this parser of that.
            #
            # Sadly, this assignment is currently ignored by the subsequent
            # call to the load() method. See also this outstanding issue:
            #     https://bitbucket.org/ruamel/yaml/issues/320/yamlload-ignores-yamlversion
            ruamel_parser.version = yaml_version

            # If assuming this file to *NOT* comply with version 1.1 of this
            # specification, ignore YAML 1.1-specific warnings. Ideally, the
            # prior assignment would suffice; since it does *NOT*, we have no
            # recourse but to ignore these warnings explicitly.
            if yaml_version != '1.1':
                context_manager = ignoring_warnings(
                    MantissaNoDotYAML1_1Warning)

        # Load and return the contents of this file with this context manager.
        with context_manager:
            return ruamel_parser.load(yaml_file)

# ....................{ SAVERS                            }....................
@type_check
def save(
    # Mandatory parameters.
    container: MappingOrSequenceTypes,
    filename: str,

    # Optional parameters.
    is_overwritable: bool = False,
) -> None:
    '''
    Save (i.e., open and write, serialize) the passed dictionary or list to the
    YAML-formatted file with the passed path via the active YAML
    implementation.

    Parameters
    ----------
    container: MappingOrSequenceTypes
        Dictionary or list to be written as the contents of this file.
    filename : str
        Absolute or relative filename of the YAML-formatted file to be saved.
    is_overwritable : optional[bool]
        Either:

        * ``True`` if this function may silently overwrite this file when this
          file already exists.
        * ``False`` if this function should instead raise an exception when
          this file already exists.

        Defaults to ``False`` for safety.
    '''

    # If this filename has no YAML-compliant filetype, log a warning.
    _warn_unless_filetype_yaml(filename)

    # With this YAML file opened for character-oriented writing...
    with iofiles.writing_chars(
        filename=filename, is_overwritable=is_overwritable) as yaml_file:
        # Fully-qualified name of the module defining this container subclass.
        container_class_module_name = objects.get_class_module_name_qualified(
            container)

        # If this container is *NOT* a "ruamel.yaml"-specific object returned
        # by a prior call to the load() function, log a non-fatal warning.
        # While this edge case does *NOT* constitute a fatal warning, it does
        # disable roundtripped preservation of comments and whitespace
        # contained in the original YAML file -- the principal motivation for
        # preferring "ruamel.yaml" versus PyYAML and friends (e.g., oyaml).
        if not container_class_module_name.startswith('ruamel.'):
            logs.log_warning(
                'Non-"ruamel.yaml" type "%s.%s" not roundtrippable.',
                container_class_module_name, container.__class__.__name__)

        # Safe roundtripping YAML parser.
        ruamel_parser = _make_ruamel_parser()

        # Save this container to this YAML file.
        ruamel_parser.dump(container, yaml_file)

# ....................{ MAKERS                            }....................
def _make_ruamel_parser() -> ruamel_yaml.YAML:
    '''
    Safe roundtripping :mod:`ruamel.yaml` parser, where:

    * "Safe" implies this parser ignores all pragmas in this YAML file
      instructing parsers to construct arbitrary Python objects, whose YAML
      syntax is of the form: ``!!python/object:module.name { ... state ... }``.
    * "Roundtripping" implies the object deserialized from this YAML file
      preserves *all* comments and whitespace of this file, such that
      re-serializing this object back to disk preserves these substrings.
    '''

    # Avoid circular import dependencies.
    from betse.lib.yaml import yamlrepr

    # Safe roundtripping YAML parser.
    ruamel_parser = ruamel_yaml.YAML(
        # Type of YAML parser to produce. Note that, by design, the
        # "rt" (i.e., roundtripping) parser is *ALWAYS* guaranteed to be safe.
        typ='rt',
    )

    # Permit this parser to roundtrip Unicode characters.
    ruamel_parser.allow_unicode = True

    # Coerce this parser into serializing nested collections as blocks rather
    # than inline "flow": e.g.,
    #
    #     # ...like this.
    #     b:
    #       c: 3
    #       d: 4
    #
    #     # ...not this.
    #     b: {c: 3, d: 4}
    ruamel_parser.default_flow_style = False

    # Locally add all representers required by this application to this parser.
    yamlrepr.add_representers(ruamel_parser.representer)

    # Return this parser.
    return ruamel_parser

# ....................{ PRIVATE ~ warners                 }....................
@type_check
def _warn_unless_filetype_yaml(filename: str) -> None:
    '''
    Log a non-fatal warning unless the passed filename has a YAML-compliant
    filetype (i.e., is suffixed by either ``.yaml`` or ``.yml``).

    Parameters
    ----------
    filename : str
        Absolute or relative filename of this file.
    '''

    # Filetype of this file.
    filetype = pathnames.get_filetype_undotted_or_none(filename)

    # If this filetype is *NOT* YAML-compliant...
    if filetype not in YAML_FILETYPES:
        # If this file has *NO* filetype, log an appropriate warning.
        if filetype is None:
            logs.log_warning('YAML file "%s" has no filetype.', filename)
        # Else, this file has a filetype. Log an appropriate warning.
        else:
            logs.log_warning(
                'YAML file "%s" filetype "%s" neither "yaml" nor "yml".',
                filename, filetype)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
CLI-specific functional tests exercising the full simulation solver.
'''

# ....................{ TESTS                              }....................
def test_cli_sim_full_noecm(betse_cli_sim: 'CLISimTester') -> None:
    '''
    Functional test exporting all available exports (e.g., CSVs, plots,
    animations) with all simulation features required by these exports,
    including the full solver but excluding extracellular spaces.

    Parameters
    ----------
    betse_cli_sim : CLISimTester
        Object running BETSE CLI simulation subcommands.
    '''

    # Enable all exports and features required by these exports, excluding ECM.
    betse_cli_sim.sim_state.config.enable_solver_full_exports_noecm()

    # Test all default simulation-specific subcommands with this configuration.
    betse_cli_sim.run_subcommands_try()


def test_cli_sim_full_ecm(betse_cli_sim: 'CLISimTester') -> None:
    '''
    Functional test exporting all available exports (e.g., CSVs, plots,
    animations) with all simulation features required by these exports,
    including both the full solver and extracellular spaces.

    Parameters
    ----------
    betse_cli_sim : CLISimTester
        Object running BETSE CLI simulation subcommands.
    '''

    # Enable all exports and features required by these exports, including ECM.
    betse_cli_sim.sim_state.config.enable_solver_full_exports_ecm()

    # Test all default simulation-specific subcommands with this configuration.
    betse_cli_sim.run_subcommands_try()


def test_cli_sim_full_vg_ions(betse_cli_sim: 'CLISimTester') -> None:
    '''
    Functional test simulating all voltage-gated ion channels (e.g., sodium,
    potassium) *and* simulation features required by these channels, including
    the full solver.

    Parameters
    ----------
    betse_cli_sim : CLISimTester
        Object running BETSE CLI simulation subcommands.
    '''

    # Enable all voltage-gated ion channels and features required by these
    # channels.
    betse_cli_sim.sim_state.config.enable_solver_full_vg_ions()

    # Test all simulation-specific subcommands *EXCLUDING* plotting subcommands
    # (which other tests already exercise) with this configuration.
    betse_cli_sim.run_subcommands_sim()



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **resource** (i.e., arbitrary file or directory contained within an
arbitrary package, regardless of whether that file or directory physically
exists *or* is compressed within an EGG-like archive file and therefore only
exists in the abstract) facilities.

See Also
----------
https://setuptools.readthedocs.io/en/latest/pkg_resources.html#resourcemanager-api
    Official :mod:setuptools documentation for the "ResourceManager API."
'''

#FIXME: Optimize this submodule to leverage the new "importlib.resources" API
#introduced with Python >= 3.7, known to be substantially faster than the
#outdated "pkg_resources" API leveraged here. Note that, since
#"importlib.resources" has been backported to Python >= 3.4, we should:
#
#* Add the backported "importlib.resources" package (...whatever they elected
#  to name it) as a new optional dependency of BETSE.
#* If "importlib.resources" is importable, prefer that over "pkg_resources" in
#  *ALL* functions defined below.

# ....................{ IMPORTS                           }....................
from betse.util.type.types import type_check
from setuptools import pkg_resources
from setuptools.pkg_resources import Requirement

# ....................{ TESTERS                           }....................
@type_check
def is_dir(module_name: (str, Requirement), dirname: str) -> bool:
    '''
    ``True`` only if the resource whose pathname is the concatenation of the
    following strings is an existing directory (*in order*):

    #. The absolute pathname of the top-level directory containing the module or
       requirement with the passed name (e.g.,
       ``/home/hiroprotagonist/betse/betse/``).
    #. The directory separator for the current platform (e.g., ``/`` on
       POSIX-compatible platforms).
    #. The passed relative pathname (e.g., ``data/yaml/sim_config.yaml``).

    Parameters
    ----------
    module_name : str, Requirement
        Either the fully-qualified name of the module *or* the
        :mod:`setuptools`- specific requirement this resource is relative to.
    dirname : str
        ``/``-separated pathname relative to this module or requirement. This
        pathname *must* be POSIX-compliant and hence separated with the POSIX-
        specific ``/`` directory separator rather than the Windows-specific
        ``\\`` directory separator (e.g., by calling ``'\\'.join()`` rather
        than either :func:`betse.util.path.paths.join` or
        :func:`os.path.join`).

    Returns
    ----------
    bool
        ``True`` only if this resource is a directory.
    '''

    # One-liners for fast and furious vengeance.
    return pkg_resources.resource_isdir(module_name, dirname)

# ....................{ GETTERS                           }....................
@type_check
def get_pathname(module_name: (str, Requirement), pathname: str) -> str:
    '''
    Absolute path of the resource whose path the pathname of the module or
    requirement with passed name joined with the passed relative pathname is an
    existing directory.

    Caveats
    ----------
    If this resource resides in an EGG-like archive file, this function
    silently extracts this resource and all resources transitively required by
    this resource to a temporary :mod:`setuptools`-specific directory *before*
    returning the absolute path of this resource within this directory.

    To quote the `ResourceManager API`_ documentation:

         If the resource is in an archive distribution (such as a zipped egg),
         it will be extracted to a cache directory, and the filename within the
         cache will be returned. If the named resource is a directory, then all
         resources within that directory (including subdirectories) are also
         extracted. If the named resource is a C extension or "eager resource"
         (see the setuptools documentation for details), then all C extensions
         and eager resources are extracted at the same time.

    :: _`ResourceManager API`:
       https://setuptools.readthedocs.io/en/latest/pkg_resources.html#resourcemanager-api

    Parameters
    ----------
    module_name : str, Requirement
        Either the fully-qualified name of the module *or* the
        :mod:`setuptools`-specific requirement this resource is relative to.
    pathname : str
        ``/``-separated pathname relative to this module or requirement. This
        pathname *must* be POSIX-compliant and hence separated with the POSIX-
        specific ``/`` directory separator rather than the Windows-specific
        ``\\`` directory separator (e.g., by calling ``'\\'.join()`` rather
        than either :func:`betse.util.path.paths.join` or
        :func:`os.path.join`).

    Returns
    -------
    str
        Absolute path of this resource.
    '''

    # One-liners for fast and furious convenience.
    return pkg_resources.resource_filename(module_name, pathname)



#!/usr/bin/env python3
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Abstract base classes of all channel classes.
'''

# ....................{ IMPORTS                            }....................
from abc import ABCMeta, abstractmethod
import numpy as np
from betse.science import sim_toolbox as stb
from betse.science.math import toolbox as tb

# ....................{ BASE                               }....................
class ChannelsABC(object, metaclass=ABCMeta):
    '''
    Abstract base class of all channel classes.

    Attributes
    ----------
    '''
    @abstractmethod
    def init(self, vm, cells, p, targets = None):
        '''
        Runs the initialization sequence for a voltage gated ion channel.
        '''

        pass



    @abstractmethod
    def run(self, vm, p):
        '''
        Runs the voltage gated ion channel.
        '''
        pass

    def update_mh(self, p, time_unit = 1e3):
        """
        Updates the 'm' and 'h' gating functions of the channel model for
        standard Hodgkin-Huxley formalism channel models.

        """

        # Update channel state using RK4:------------------------------------
        # dm = tb.RK4(lambda m: (self._mInf - self.m) / (self._mTau))
        # dh = tb.RK4(lambda h: (self._hInf - self.h) / (self._hTau))
        #
        # self.m += dm(self.m, p.dt)
        # self.h += dh(self.h, p.dt)

        # Update channel state using semi-Implicit Euler method:-------------------
        dt = p.dt*self.time_unit

        self.m = (self._mTau*self.m + dt*self._mInf)/(self._mTau + dt)
        self.h = (self._hTau*self.h + dt*self._hInf)/(self._hTau + dt)

    def update_ml(self, p, time_unit = 1e3):
        """
        An iterative time-step updating method for the
        time-dependent Morris-Lecar formalism of voltage-gated
        ion channels (a variant of Hodgkin-Huxley)
        :param p: Parameters
        :param time_unit:
        :return:
        """

        dt = p.dt * self.time_unit
        self.m = (self.m + (dt * self.Phi * self._mInf / self._mTau)) / (1 + ((dt * self.Phi) / self._mTau))



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level wrappers around Pympler, an optional runtime dependency performing
memory profiling of Python applications.
'''

#FIXME: Consider donating the get_size_profile() function back to Pympler, which
#could *REALLY* benefit from a human-readable CLI-based object graph profile.

# ....................{ IMPORTS                           }....................
from betse.util.type import types
from betse.util.type.types import type_check, NumericSimpleTypes

# ....................{ FORMATTERS                        }....................
_SIZE_FORMATTER = r'{:.04f} MiB'
'''
Format specifier converting the size in bytes interpolated into this string
with the :func:`format` builtin into a human-readable string.
'''


_TYPE_FORMATTER = r'<{.__class__.__name__}>: '
'''
Format specifier converting an object type interpolated into this string with
the :func:`format` builtin into a human-readable string.
'''

# ....................{ GETTERS                           }....................
@type_check
def get_size(obj: object, size_divisor: NumericSimpleTypes = 1) -> (
    NumericSimpleTypes):
    '''
    Recursive in-memory size of the passed object, calculated with the external
    :mod:`pympler` package if importable _or_ raising an exception otherwise.

    This size is formally defined as the sum of:

    * The flat size of this object, defined as either:

      * If the type of this object provides the `__sizeof__()` special method,
        the return value of the :func:`sys.getsizeof` function passed this
        object.
      * Else, the sum of:

        * The basic size of this object, defined as the value of the
          undocumented `__basicsize__` attribute of the type of this object.
          Although undocumented, this appears to be the size of this object
          when **empty** (i.e., when this object contains no elements). For
          garbage-collected objects, this size "...includes the overhead for
          Python’s garbage collector (GC) as well as the space needed for
          refcounts (used only in certain Python builds)."
        * The item size of this object, defined as the value of the
          undocumented `__itemsize__` attribute of the type of this object.
          Although undocumented, this size appears to be:

          * If this object is a container, the sum of the sizes of the
            references referring to all objects contained by this object but
            _not_ the sizes of these objects themselves.
          * If this object is _not_ a container, 0.

    * The recursive size of each **referent** (i.e., object referred to by an
      instance variable) of this object, visited recursively up to some finite
      recursion depth. The size of a referent:

      * Visitable only by recursion exceeding this depth is silently ignored.
      * Visited multiple times is incorporated only once. Each subsequent
        visitation of this referent is silently ignored.

    Caveats
    ----------
    **This function requires the external :mod:`pympler` package,** an optional
    runtime dependency of this application. As the internal complexity of the
    :mod:`pympler.asizeof` submodule suggests, recursively calculating object
    sizes in a cross-platform and -interpreter portable manner generically
    applicable to _all_ possible objects is highly non-trivial. Since the only
    function defined by the stdlib to calculate object sizes (i.e.,
    :func:`sys.getsizeof`) behaves non-recursively, recursively calculating
    these sizes would effectively require wrapping that low-level function with
    a brute-force depth- or breadth-first search (DFS).

    Doing so would effectively reimplement the majority of the
    :mod:`pympler.asizeof` submodule, a lamentable time waste violating the
    Don't Repeat Yourself (DRY) principle in the best case and a futile failure
    in the worst. To preserve sanity, :mod:`pympler` is imported instead.

    Parameters
    ----------
    obj : object
        Object to iterate the variable sizes of.
    size_divisor: optional[NumericSimpleTypes]
        Number (i.e., integer, float) by which to divide the size in bytes of
        each variable of this object, hence converting from sizes in bytes to
        sizes in a larger denomination (e.g., :data:`betse.util.type.ints.KiB`,
        yielding sizes in kibibytes). Defaults to 1, yielding sizes in bytes.

    Returns
    ----------
    int
        Recursive in-memory size of this object in bytes, divided by the passed
        divisor if any.

    Raises
    ----------
    BetseLibException
        If the optional runtime dependency Pympler is unsatisfied.
    '''

    # Avoid circular import dependencies.
    from betse.lib import libs

    # If Pympler is unavailable, raise an exception.
    libs.die_unless_runtime_optional('pympler')

    # Import Pympler *AFTER* validating its availability.
    from pympler import asizeof

    # Return the in-memory recursive size of this object, converted from bytes
    # to another denomination by the passed divisor.
    return asizeof.asizeof(obj) / size_divisor


#FIXME: Call this method as a new functional test detecting erroneously large
#attributes in these three principal objects. To do so, determine the largest
#attribute currently produced in these objects for the minified world
#environment used for all functional tests. Then raise an exception if any
#attribute of these objects is greater than or equal to 10 times this
#expected maximum.

@type_check
def get_sizes_vars(obj: object, *args, **kwargs) -> tuple:
    '''
    2-tuple `(obj_size, vars_custom_size)` of this object's total recurusive
    size _and_ sequence of 2-tuples of the name and total recursive size of
    each **non-builtin variable** (i.e., variable whose name is *not* both
    prefixed and suffixed by `__`) bound to the passed object (in descending
    order of size), calculated with the external :mod:`pympler` package if
    importable *or* raising an exception otherwise.

    Parameters
    ----------
    obj : object
        Object to iterate the variable sizes of.

    All remaining positional and keyword arguments are passed as is to the
    :func:`get_size` function.

    Returns
    ----------
    (int, SequenceTypes)
        2-tuple `(obj_size, vars_custom_size)`, where

        * `obj_size` is this object's **total recursive size** (i.e., summation
          of the total non-recursive sizes of all non-builtin variables bound
          to this object).
        * `vars_custom_size` is a sequence of 2-tuples `(var_name, var_size)`
          of the name and recursive in-memory size of each variable bound to
          this object (_in descending order of size_).

    Raises
    ----------
    BetseLibException
        If the optional runtime dependency Pympler is unsatisfied.
    Exception
        If any variable of this object is a property whose
        :func:`property`-decorated method raises an exception.

    See Also
    ----------
    :func:`get_size`
        Further details on object size calculation.
    :func:`iter_vars_custom`
        Further details on non-builtin object variables.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.iterable import itersort
    from betse.util.type.numeric import numerics
    from betse.util.type.obj import objiter

    # List of all 2-tuples to be returned.
    vars_size = []

    # For the name and value of each non-builtin variable of this object (in
    # arbitrary order)...
    for var_name, var_value in objiter.iter_vars_custom(obj):
        # Recursive in-memory size of this object.
        var_size = get_size(obj=var_value, *args, **kwargs)

        # Append a 2-tuple of this variable's name and size.
        vars_size.append((var_name, var_size))

    # Total recursive size of this object, totalized over the second element of
    # each 2-tuple in this list.
    obj_size = numerics.sum_by_index(iterable=vars_size, subiterable_index=1)

    # This list sorted on the second element of each such 2-tuple (i.e., each
    # variable's size) in descending order.
    vars_size_sorted = itersort.sort_by_index_descending(
        iterable=vars_size, subiterable_index=1)

    # Return the expected 2-tuple.
    return obj_size, vars_size_sorted

# ....................{ GETTERS ~ profile                 }....................
#FIXME: Revise docstring, removing reference to get_sizes_vars() in particular.
@type_check
def get_size_profile(
    obj: object,
    *args,
    line_indent: str = '',
    vars_depth: int = 1,
    vars_max: int = None,
    **kwargs
) -> str:
    '''
    Human-readable synopsis of the size of the passed object with all remaining
    arguments passed as is to the :func:`get_sizes_vars` function, calculated
    with the external :mod:`pympler` package if importable _or_ raising an
    exception otherwise

    Specifically, this function returns a string tabulating:

    * The total recursive in-memory size in mebibytes of this object.
    * The names and total recursive in-memory sizes in mebibytes of all
      **non-builtin variables** (i.e., variables whose names are *not* both
      prefixed and suffixed by `__`) bound to this object (in descending order
      of size).

    Motivation
    ----------
    Although Pympler does define numerous classes and callables for profiling
    arbitrary objects, the resulting output is non-human-readable in the best
    case and ambiguous in the worst case. In either case, such output is mostly
    unusable. For example:

        >>> from betse.util.type.obj import objsize
        >>> from pympler.asizeof import asizesof
        >>> class ObjectLessons(object):
        ...     def __init__(self):
        ...         self.lessons = [
        ...             'Easter Island', 'St. Matthew Island', 'Greenland Norse',]
        ...         self.lesson_count = len(self.lessons)
        >>> object_lessons = ObjectLessons()
        >>> asizesof(object_lessons, stats=2.0)
        592 bytes:  <__main__.ObjectLessons object at 0x7fe937595208>
        592 bytes
        8 byte aligned
        8 byte sizeof(void*)
        1 object given
        10 objects sized
        10 objects seen
        3 recursion depth

        5 profiles:  total (% of grand total), average, and largest flat size:  largest object
        5 class str objects:  320 (54%), 64, 72:  'St. Matthew Island' leng 19!
        1 class dict object:  96 (16%), 96, 96:  {'lesson_count': 3, 'lessons': ['Easte..... Matthew Island', 'Greenland Norse']} leng 0
        1 class list object:  88 (15%), 88, 88:  ['Easter Island', 'St. Matthew Island', 'Greenland Norse'] leng 7!
        1 class __main__.ObjectLessons object:  56 (9%), 56, 56:  <__main__.ObjectLessons object at 0x7fe937595208>
        1 class int object:  32 (5%), 32, 32:  3 leng 1!

    Critically, note that:

    * The total size of the `ObjectLessons.lessons` list is actually 288 bytes
      rather than the significantly smaller 88 bytes reported above.
      Specifically:

      * The flat size of this list is 88 bytes, which Pympler erroneously
        reports to be the total size of this list.
      * The total flat size of all strings contained in this list is 200 bytes.
        Ergo, the total size of this list is 288 bytes -- as verified by
        explictly passing this list to the
        :func:`pympler.asizeof.asizeof` function, which then returns 288.

    * The total size of each variable of the passed `ObjectLessons` instance is
      *not* reported. Only the total flat size of each type of object
      recursively visitable from this instance is reported, which is useless.

    Until Pympler addresses both concerns, this function profiles the memory
    consumption of arbitrary objects in a human-readable manner.

    Parameters
    ----------
    obj : object
        Object to synopsize the size of.
    line_indent: optional[str]
        String to prefix each line of the returned string by, typically
        consisting of zero or more space characters. Defaults to the empty
        string.
    vars_depth: optional[int]
        Maximum depth of the recursion tree induced by calling this function.
        If this depth is:

        * Zero, a single-line synopsis of the total size of this object is
          returned *without* totalling the sizes of any variables bound to this
          object. This depth is appropriate for C-based objects (i.e., objects
          whose classes are either primitive builtins or defined by external C
          extensions), whose details are usually inconsequential when profiling
          the space consumed by pure-Python objects. Since depth is necessarily
          positive, recursion "bottoms out" at this depth.
        * Non-zero, a single-line synopsis of the total size of this object
          appended by the total sizes of all variables bound to this object is
          recursively returned. Specifically, for each such variable, this
          function recursively calls itself with:

          * `obj` parameter equal to the value of that variable.
          * `line_indent` parameter appended by two spaces.
          * `vars_depth` parameter decremented by one.
          * `vars_max` parameter divided by the total number of variables bound
            to this object.

        Defaults to 1, in which case a single-line synopsis of the total size
        of this object appended by the total sizes of all variables directly
        bound to this object is non-recursively returned.
    vars_max : optional[int]
        Maximum number of the largest non-builtin variables bound to this object
        to synopsize the sizes of. Defaults to `None`, in which case this
        function unconditionally synopsizes the sizes of *all* non-builtin
        variables bound to this object (regardless of size).

    All remaining positional and keyword arguments are passed as is to the
    :func:`get_size` function.

    Returns
    ----------
    str
        Human-readable synopsis of the size of this object.

    Raises
    ----------
    BetseLibException
        If the optional runtime dependency Pympler is unsatisfied.
    Exception
        If any variable of this object is a property whose
        :func:`property`-decorated method raises an exception.

    See Also
    ----------
    :func:`iter_vars_custom_size`
        Further details on object size calculation.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.iterable import itersort
    from betse.util.type.numeric.ints import MiB
    from betse.util.type.obj import objiter, objtest
    from betse.util.type.text.string import strjoin, strs

    #FIXME: Convert the current "kwargs['size_divisor']" into a human-readable
    #denomenation (e.g., "GiB", "KiB", "B"). Since this is both non-trivial and
    #currently *NOT* required, we adopt a superior lazy approach.

    # If the caller failed to specify a size divisor, default to a divisor
    # denominating sizes in mebibytes for readability.
    kwargs.setdefault('size_divisor', MiB)

    # Set of the unique identifiers of all objects previously passed to a prior
    # recursive call by the closure defined below, preventing erroneous
    # reinspection of these objects.
    obj_ids = set()

    #FIXME: Revise docstring.

    # For efficiency, this function is intentionally *NOT* type-checked.
    def _get_size_profile_and_size(
        obj: object,
        line_indent: str,
        vars_depth: int,
        vars_max: int,
    ) -> tuple:
        '''
        Human-readable synopsis of the size of the passed object with all
        remaining arguments passed as is to the :func:`get_sizes_vars` function,
        calculated with the external :mod:`pympler` package if importable *or*
        raising an exception otherwise.

        Specifically, this function returns a string tabulating:

        * The total recursive in-memory size in mebibytes of this object.
        * The names and total recursive in-memory sizes in mebibytes of all
          **non-builtin variables** (i.e., variables whose names are *not* both
          prefixed and suffixed by `__`) bound to this object (in descending
          order of size).

        Parameters
        ----------
        This function has the same signature with respect to passed parameters
        as the :func:`get_size_profile` function.

        Returns
        ----------
        (str, NumericOrNoneTypes)
            2-tuple `(obj_size_profile, obj_size)`, where:

            * `obj_size_profile` is a recursive size profile of this object.
            * `obj_size` is the total recursive size of this object in bytes if
              this object has *not* already been passed to a prior recursive
              call of this closure or ``None`` otherwise. Note that, in this
              case, ``obj_size_profile`` is a relevant human-readable string
              indicating this circularity in the object graph.
        '''

        # If this object has already been passed to a prior recursive call of
        # this closure, skip this object and notify the caller of our having
        # done so by returning the empty 2-tuple.
        if id(obj) in obj_ids:
            # Total recursive size of this object, zeroed to prevent parent
            # recursive calls from erroneously re-summing with this size.
            obj_size = 0

            # Notify the end user of this object graph circularity.
            size_profile = (
                _TYPE_FORMATTER + '(duplicate object already sized above)'
            ).format(obj)

            # Return the expected 2-tuple up the recursive call stack.
            return size_profile, obj_size

        # Else, this object has *NOT* yet been passed to a prior recursive call
        # of this closure. To prevent recursive calls below from erroneously
        # reinspecting this object, record this object's unique identifier.
        obj_ids.add(id(obj))

        # True only if this function is immediately returning *WITHOUT*
        # recursivily calling itself below.
        #
        # Specifically, if any of the following conditions apply, halt this
        # recursive subtree by non-recursively returning a simple size profile:
        #
        # * This is the bottom-most recursive call in the current
        #   subtree of recursive calls.
        # * The parent recursive call requested that this child recursive call
        #   profile only this object itself (rather than both this object and
        #   all or some variables bound to this object).
        # * No variables are bound to this object.
        # * This object is C-based (i.e., is either builtin or defined by a C
        #   extension).
        #
        # Note that the second such test (i.e., "vars_max < 1") is technically
        # non-essential, as the list truncation performed below already
        # implicitly handles this edge cases without error. Doing so is highly
        # inefficient, however, and hence best avoided if feasible. It is
        # feasible, so it is.
        #
        # If this object is C-based, recursively introspecting this object
        # would be trivially feasible. Python code makes no distinction between
        # pure-Python and C-based objects. Indeed, differentiating between the
        # two with Python code is non-trivial! That said, recursively
        # introspecting C-based objects would be largely pointless; size
        # profiles are principally intended to profile the space consumed by
        # variables bound to application-specific pure-Python objects rather
        # than those bound to external C-based objects.
        is_nonrecursive = (
            vars_depth == 0 or vars_max < 1 or objtest.is_c_based(obj))

        # If this function is possibly recursively calling itself below...
        if not is_nonrecursive:
            # Tuple of 2-tuples "(var_name, var_value)" for each non-builtin
            # variable bound to this object, permitting the number of 2-tuples
            # yielded by this generator to be summed below.
            #
            # For efficiency, this inefficient operation is performed only if
            # this function is possibly recursively calling itself below. If
            # this function is *NOT* doing so, this operation is avoided.
            vars_name_value = tuple(objiter.iter_vars_custom(obj))

            # Number of such variables.
            vars_count = len(vars_name_value)

            # If no variables are bound to this object, return immediately.
            is_nonrecursive = not vars_count

        # If this function is returning *WITHOUT* recursivily calling itself...
        if is_nonrecursive:
            # Total recursive size of only this object.
            obj_size = get_size(obj, *args, **kwargs)

            # Total recursive size profile of only this object.
            size_profile = (
                _TYPE_FORMATTER + _SIZE_FORMATTER).format(obj, obj_size)

            # Return the expected 2-tuple up the recursive call stack.
            return size_profile, obj_size
        # Else, recursively return a size profile of both this object *AND* all
        # or some variables bound to this object. By the above logic, this
        # object is guaranteed to contain at least one such variable.

        # Human-readable string suffixing this synopsis if any or empty.
        size_profile_suffix = ''

        # Substring prefixing each line of variable-specific synopsis, indented
        # to the right by two additional spaces.
        var_line_indent = line_indent + '  '

        # Maximum depth of the recursion tree induced by calling this function
        # recursively below for each variable bound to this object. To ensure
        # that recursion bottoms out, the current maximum is decremented by one.
        var_vars_depth = vars_depth - 1

        # Maximum number of variables bound to each variable bound to this
        # object to synopsize the sizes of if any or None otherwise.
        var_vars_max = vars_max

        # If such a maximum is requested, distribute this maximum amongst all
        # available variables, rounding down to the nearest integer.
        #
        # Note that, if the number of such variables is larger than this
        # maximum, this maximum will be reduced to zero. This valid edge case
        # is explicitly handled by the conditional above.
        if var_vars_max is not None:
            var_vars_max = round(vars_max / vars_count)

        # List of 2-tuples "(var_size_profile, var_size)" for each variable
        # bound to this object, where:
        #
        # * "var_size_profile" is a recursive size profile of this variable.
        # * "var_size" is the total recursive size of this variable in bytes.
        #
        # Associating each profile with its size is required to subsequently
        # sort and possibly discard profiles by size.
        vars_size_profile = []

        # Total recursive size of this object, incremented by iteration below.
        obj_size = 0

        #FIXME: Consider left-aligning the sizes interpolated by this iteration.
        #FIXME: Consider interpolating the truncation of the value of each
        #variable in this iteration.

        # For the name and size of each of this object's variables...
        for var_name, var_value in vars_name_value:
            # Size profile and size of this variable, computed recursively.
            var_size_profile, var_size = _get_size_profile_and_size(
                obj=var_value,
                line_indent=var_line_indent,
                vars_depth=var_vars_depth,
                vars_max=var_vars_max,
            )

            # Prefix this variable's size profile by indentation and its name.
            var_size_profile = '{}{} {}'.format(
                var_line_indent, var_name, var_size_profile)

            # Append this size and size profile to this list as a 2-tuple.
            vars_size_profile.append((var_size_profile, var_size))

            # Increment the size of this parent object by this child variable.
            obj_size += var_size

        # This list sorted on the second element of each such 2-tuple (i.e.,
        # each variable's size) in descending order.
        vars_size_profile_sorted = itersort.sort_by_index_descending(
            iterable=vars_size_profile, subiterable_index=1)

        # If a maximum number of the largest variables is requested *AND* the
        # number of variables bound to this object exceeds this maximum, ignore
        # all variables of smaller size. To ensure the synopsis suffix set
        # below has proper indentation, do so *AFTER* increasing this level.
        if vars_max is not None and vars_count > vars_max:
            # print('!!Truncating!!')
            # Validate this number to be a positive integer.
            assert types.is_int_positive(vars_max), (
                types.assert_not_int_positive(vars_max))

            # Truncate this list to this maximum number of the first 2-tuples.
            vars_size_profile_sorted = vars_size_profile_sorted[:vars_max]

            # Append an ellipses to this synopsis, notifying users that
            # additional variables exist but were truncated.
            size_profile_suffix = '{}...'.format(var_line_indent)

        # print('vars_size_profile_sorted: {}'.format(vars_size_profile_sorted))
        # print('vars_size_profile_sorted:')
        # for var_size_profile_sorted, var_size_sorted in (
        #     vars_size_profile_sorted):
        #     print('  str: "{}"; int: {}'.format(var_size_profile_sorted, var_size_sorted))

        #FIXME: For efficiency, define a new "_TYPE_SIZE_FORMATTER" global
        #concatenating "_TYPE_FORMATTER" and "_SIZE_FORMATTER". Then remove both
        #"_TYPE_FORMATTER" and "_SIZE_FORMATTER", which will then be extraneous.

        # Human-readable synopsis of...
        size_profile = (
            # The size of this object.
            (_TYPE_FORMATTER + _SIZE_FORMATTER).format(obj, obj_size) +

            # The size of each variable bound to this object.
            strjoin.join_by_index(
                iterable=vars_size_profile_sorted, subiterable_index=0) +

            # The substring suffixing this entire synopsis if any.
            size_profile_suffix +

            # A blank line, improving readability and aesthetics.
            '\n'
        )

        # Return the expected 2-tuple up the recursive call stack.
        return size_profile, obj_size

    # Size profile recursively computed from this object. The second element of
    # the tuple returned by this function is intended for use *ONLY* by
    # recursive calls by this function of itself and is hence ignorable here.
    size_profile, _ = _get_size_profile_and_size(
        obj,

        # Prefix the passed substring prefixing every output line by a newline,
        # conveniently simplifying the implementation of _get_size_profile().
        line_indent='\n' + line_indent,

        # Pass all other passed parameters as is.
        vars_depth=vars_depth,
        vars_max=vars_max,
    )

    # Munge this profile as follows:
    #
    # * Prefix this profile by this indentation. For inscrutable recursion
    #   reasons, the first element of the tuple returned by this function is a
    #   string guaranteed *NOT* to be prefixed by this indentation
    # * Remove all suffixing whitespace -- namely, the trailing blank line
    #   resulting again from inscrutable recursion reasons.
    size_profile = line_indent + strs.remove_whitespace_suffix(size_profile)

    # Return this munged profile.
    return size_profile



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level operating system (OS)-specific display facilities.
'''

#FIXME: Submit a Stackoverflow answer encapsulating this logic. The
#osx.is_aqua() function in particular would be useful to a wide audience.

# ....................{ IMPORTS                           }....................
from betse.util.io.log import logs
from betse.util.type.decorator.decmemo import func_cached
from betse.util.type.iterable.mapping.mapcls import OrderedArgsDict
from betse.util.type.types import type_check

# ....................{ GLOBALS                           }....................
_is_headless_forced = None
'''
``True`` or ``False`` only if the :func:`set_headless` function (which
overrides the detection performed by the :func:`is_headless` function of
whether the active Python interpreter is running headless) has been called at
least once, in which case all subsequent calls to the :func:`is_headless`
function returns this boolean rather than performing that detection.

Defaults to ``None``, in which case the :func:`is_headless` function performs
such detection rather than returning the value of this boolean.
'''

# ....................{ TESTERS                           }....................
@func_cached
def is_dpi_scaling() -> bool:
    '''
    ``True`` only if the current platform natively supports high-DPI scaling.

    Specifically, if the current platform is:

    * Linux *and* the active Python interpreter is running under a Wayland
      compositor, this function returns ``True``.
    * macOS, this function returns ``True``.
    * Windows >= 10, this function returns ``True``. Technically, different
      versions of Windows 10 allow end users to conditionally disable high-DPI
      scaling for different use cases. Since detecting these versions and
      settings in pure-Python is effectively infeasible, this function
      simplistically assumes all installations of Windows 10 and newer to
      unconditionally enable high-DPI scaling.

    All other platforms are assumed to *not* natively support high-DPI scaling.
    This includes both the Linux-centric X11 display server *and* all versions
    of Windows preceding Windows 10 (i.e., Windows <= 8).
    '''

    # Avoid circular import dependencies.
    from betse.util.os.brand import linux, macos, windows

    # Return true only if this platform is either...
    return (
        # A Linux distribution running Wayland *OR*...
        linux.is_wayland() or
        # macOS *OR*...
        macos.is_aqua() or
        # Windows >= 10.
        windows.is_version_10_or_newer()
    )

# ....................{ TESTERS ~ head(full|less)         }....................
# Note that the following public testers *CANNOT* be memoized (e.g., via the
# @func_cached decorator), as the set_headless() function allows callers to
# externally modify the return values of these testers at any time. Instead,
# the private _is_headless() tester (to which these public testers defer)
# internally caches the costly detection it performs on the first call to that
# tester and then subsequently returns that cached boolean.

def is_headfull() -> bool:
    '''
    ``True`` only if the active Python interpreter is running **headfull**
    (i.e., with access to a GUI display, the common case when running under a
    conventional desktop, laptop, or tablet device).
    '''

    return not is_headless()     # Makes sense.


def is_headless() -> bool:
    '''
    ``True`` only if the active Python interpreter is running **headless**
    (i.e., with *no* access to a GUI display, often due to running remotely
    over an SSH-encrypted connection supporting only CLI input and output).

    See Also
    ----------
    :func:`_is_headless`
        Further details.
    '''

    # If the set_headless() function has been called at least once, return the
    # boolean passed to the most recent call to that function.
    if _is_headless_forced is not None:
        return _is_headless_forced
    # Else, the set_headless() function has *NOT* been called at least once. In
    # this case, intelligently detect whether this process is headless or not.

    # Return true only if this interpreter is running headless.
    return _is_headless()


@func_cached
def _is_headless() -> bool:
    '''
    ``True`` only if the active Python interpreter is running **headless**
    (i.e., with *no* access to a GUI display, often due to running remotely
    over an SSH-encrypted connection supporting only CLI input and output).

    Specifically, this function returns:

    * If the :func:`set_headless` function has been called at least once, the
      boolean passed to the most recent call to that function. Ergo, that
      function overrides the intelligent detection performed by this function.
    * Else, ``True`` only if none of the following conditions apply:

      * The current platform is Microsoft Windows. While a small subset of
        server-specific variants of Windows can and often are run headless
        (e.g., Windows Nano Server), there appears to be no known means of
        reliably distinguishing a headless from headfull Windows environment in
        pure Python. For safety, we assume the latter.
      * The current platform is not Microsoft Windows (and hence is
        POSIX-compatible) *and* either:

        * The ``${DISPLAY}`` environment variable required by the
          POSIX-compatible X11 display server is set under the current
          environment (typically due to being inherited from the parent
          environment). Note that *all* POSIX-compatible platforms of interest
          including macOS support this server.
        * The current platform is macOS *and* the :func:`macos.is_aqua` returns
          ``True``.
        * The current platform is Linux *and* either:

          * The ``${MIR_SOCKET}`` environment variable required by the
            Linux-specific Mir display server is set under this environment.
          * The ``${WAYLAND_DISPLAY}`` environment variable required by the
            Linux-specific Wayland display server is set under this
            environment.
    '''

    # Avoid circular import dependencies.
    from betse.util.os.brand import linux, macos, posix, windows

    # The active Python interpreter is headfull if and only if either...
    is_os_headfull = (
        # This is Windows, in which case this interpreter is usually headfull.
        # While certain server-specific variants of Windows can and often are
        # run headless (e.g., Windows Nano Server), there appears to be no
        # known means of reliably distinguishing a headless from headfull
        # Windows environment in pure Python. For safety, assume the latter.
        windows.is_windows() or

        # Else, this is a POSIX-compatible platform.
        #
        # Since all POSIX-compatible platforms of interest support the popular
        # X11 display server, detect this server first.
        posix.is_x11() or

        # Else, all possible alternative display servers specific to the
        # current platform *MUST* be iteratively tested for.
        #
        # If Linux, the only remaining display servers are Mir and Wayland.
        (linux.is_linux() and (linux.is_wayland() or linux.is_mir())) or

        # If macOS, the only remaining display server is Aqua.
        (macos.is_macos() and macos.is_aqua())

        # Else, this platform is unrecognized. For safety, this platform is
        # assumed to be headless.
    )

    # Return true only if this interpreter is *NOT* headfull.
    #
    # Note that the "is_os_headfull" boolean intentionally contains the core
    # detection logic, as detecting headfull environments is fundamentally
    # more intuitive than detecting the converse.
    return not is_os_headfull

# ....................{ SETTERS                           }....................
@type_check
def set_headless(is_headless: bool) -> None:
    '''
    Set whether the active Python interpreter is running **headless** (i.e.,
    with *no* access to a GUI display) or not, explicitly overriding the
    implicit detection performed by the :func:`is_headfull` function of whether
    this interpreter actually is running headless or not.

    Parameters
    ----------
    is_headless : bool
        ``True`` only if the active Python interpreter is to be treated as if
        it were effectively running headless - regardless of whether it is.
    '''

    # Enable this global to be locally set.
    global _is_headless_forced

    # If coercing headless operation, log this coercion.
    if is_headless:
        logs.log_debug('Forcing headless operation...')
    # Else, headfull operation is being coerced. In this case...
    else:
        # Log this coercion.
        logs.log_debug('Forcing headfull operation...')

        # If the current environment is detected to be headless, log a
        # non-fatal warning. While an exception could also be raised, our
        # detection heuristic is known to be imperfect.
        if _is_headless():
            logs.log_warning(
                'Headless environment detected! '
                'Forcing headfull operation under a headless environment '
                'typically raises silent segmentation faults '
                'and hence is unsupported.'
            )

    # Set this global to this boolean.
    _is_headless_forced = is_headless

# ....................{ GETTERS ~ metadata                }....................
def get_metadata() -> OrderedArgsDict:
    '''
    Ordered dictionary synopsizing the current display.
    '''

    # Avoid circular import dependencies.
    from betse.util.os.brand import linux, macos, posix

    # Return this dictionary.
    return OrderedArgsDict(
        'headless',    is_headless(),
        'dpi scaling', is_dpi_scaling(),
        'aqua',        macos.is_aqua(),
        'mir',         linux.is_mir(),
        'wayland',     linux.is_wayland(),
        'x11',         posix.is_x11(),
    )



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **process tester** (i.e., utility functions testing and validating
currently running platform-specific processes) facilities.
'''

# ....................{ IMPORTS                           }....................
from betse.exceptions import (
    BetseFunctionUnimplementedException,
    BetseProcessNotFoundException,
)
from betse.util.type.types import type_check

# ....................{ TESTERS ~ parent                  }....................
#FIXME: This tester is currently uncalled. Before calling this tester, the
#optional "psutil" dependency *MUST* be uncommented in "betse.metadeps".
@type_check
def is_parent_command(command_basename: str) -> bool:
    '''
    ``True`` only if the parent process of the active Python interpreter is
    running an external command with the passed basename optionally suffixed by
    a platform-specific filetype (e.g., ``.exe`` under Windows).

    Parameters
    ----------
    command_basename : str
        Basename of the command to test this process' parent process against.

    Raises
    ------
    BetseOSException
        If the optional :mod:`psutil` dependency is unimportable *and* the
        current platform is neither Linux, macOS, or Windows.
    BetseFunctionUnimplementedException
        If the optional :mod:`psutil` dependency is unimportable.
    BetseProcessNotFoundException
        If either:

        * The optional :mod:`psutil` dependency is importable *and* either:

          * The current process has *no* parent process.
          * The current process had a parent process that has since died.

    Returns
    ----------
    bool
        ``True`` only if the parent process of the active Python interpreter is
        running an external command with the passed basename.
    '''

    # Avoid circular import dependencies.
    from betse.lib import libs
    from betse.util.os import oses
    from betse.util.os.brand import linux, macos, windows

    # If the optional "psutil" dependency is importable, this dependency
    # typically implements more efficient *AND* portable solutions than all
    # alternatives and is hence preferred. In this case, defer to "psutil".
    if libs.is_runtime_optional('psutil'):
        return _is_parent_command_psutil(command_basename)
    # Else if the current platform is Linux, defer to a Linux-specific tester.
    elif linux.is_linux():
        return _is_parent_command_linux(command_basename)
    # Else if the current platform is macOS, defer to a macOS-specific tester.
    elif macos.is_macos():
        return _is_parent_command_macos(command_basename)
    # Else if the current platform is Windows, defer to a Windows-specific
    # tester.
    elif windows.is_windows():
        return _is_parent_command_windows(command_basename)

    # Else, raise an exception.
    oses.die_if_unsupported()


@type_check
def _is_parent_command_psutil(command_basename: str) -> bool:
    '''
    ``True`` only if the parent process of the active Python interpreter is
    running an external command with the passed basename, implemented in terms
    of the optional :mod:`psutil` dependency.

    See Also
    ----------
    https://stackoverflow.com/a/2241047/2809027
        StackOverflow answer strongly inspiring this implementation.
    :func:`is_parent_command`
        Further details.
    '''

    # Avoid circular import dependencies.
    from betse.lib import libs
    from betse.util.os.brand import windows
    from betse.util.path import pathnames

    # Optional "psutil" dependency.
    psutil = libs.import_runtime_optional('psutil')

    # Object encapsulating the current process.
    current_proc = psutil.Process()

    # Object encapsulating the parent process of this process if any *OR*
    # "None" otherwise.
    parent_proc = current_proc.parent()

    # If the current process has *NO* parent process, raise an exception.
    #
    # In theory, all processes except the initial "init" process should
    # *ALWAYS* have a parent process on POSIX-compatible platforms. In
    # practice, "psutil" documentation explicitly states that the
    # psutil.Process.parent() method returns "None" in edge cases. *sigh*
    if parent_proc is None:
        raise BetseProcessNotFoundException(
            'Current process {} parent not found.'.format(current_proc))
    # Else, the current process has a parent process.

    # If...
    if (
        # The current platform is Windows *AND*...
        windows.is_windows() and
        # This command basename is *NOT* suffixed by a filetype...
        not pathnames.is_filetype(command_basename)
    # Then suffix this basename by ".exe". This is required, as Windows
    # executables are *ALWAYS* suffixed by such a filetype.
    ):
        command_basename += '.exe'

    # Attempt to...
    try:
        # Dictionary of all metadata required to implement this tester.
        #
        # Note that this call:
        #
        # * Efficiently retrieves this metadata with one method call rather
        #   than inefficiently distributing this retrieval across multiple
        #   method calls.
        # * Transparently sets the values of all metadata whose retrieval
        #   raises an "AccessDenied" or "ZombieProcess" exception to "None."
        #   Sadly, the "NoSuchProcess" exception is *NOT* handled similarly and
        #   *MUST* thus be explicitly caught below.
        parent_proc_metadata = parent_proc.as_dict(
            attrs=('name', 'exe', 'cmdline'))

        # List of one or more strings comprising the command line invoking
        # this parent process *OR* "None" if retrieving this metadata
        # raised an exception above.
        parent_proc_cmdline = parent_proc_metadata['cmdline']

        # If either...
        return (
            # This command basename is this parent process' name *OR*...
            #
            # Note that this metadata typically requires the least privelages
            # on most platforms *AND* is typically the most efficient such
            # metadata to access. So, this metadata is tested first.
            command_basename == parent_proc_metadata['name'] or
            # This command basename is that of this parent process *OR*...
            command_basename == pathnames.get_basename(
                parent_proc_metadata['exe']) or
            (
                # This parent process' command line was retrievable *AND*...
                parent_proc_cmdline and
                # This command basename is that of the first item of this line.
                command_basename == pathnames.get_basename(
                    parent_proc_cmdline[0])
            )
        )
    # If this parent process died during metadata access, raise an exception.
    except psutil.NoSuchProcess as exception:
        raise BetseProcessNotFoundException(
            'Current process {} parent {} killed.'.format(
                current_proc, parent_proc)
        ) from exception


#FIXME: Implement us up given the StackOverflow answer listed below.
@type_check
def _is_parent_command_linux(command_basename: str) -> bool:
    '''
    ``True`` only if the parent process of the active Python interpreter is
    running an external command with the passed basename, implemented in terms
    of the Linux-specific **process pseudo-filesystem** (i.e., ``/proc``).

    See Also
    ----------
    https://stackoverflow.com/a/24114907/2809027
        StackOverflow answer strongly inspiring this implementation.
    :func:`is_parent_command`
        Further details.
    '''

    raise BetseFunctionUnimplementedException()


#FIXME: Research whether implementing this tester is even feasible.
@type_check
def _is_parent_command_macos(command_basename: str) -> bool:
    '''
    ``True`` only if the parent process of the active Python interpreter is
    running an external command with the passed basename, implemented in terms
    of the macOS-specific... well, who actually knows?

    See Also
    ----------
    :func:`is_parent_command`
        Further details.
    '''

    raise BetseFunctionUnimplementedException()


#FIXME: Incorporate the following brilliant pure-Python solution for parsing
#Windows process metadata without forking external processes:
#    https://stackoverflow.com/a/7110486/2809027
@type_check
def _is_parent_command_windows(command_basename: str) -> bool:
    '''
    ``True`` only if the parent process of the active Python interpreter is
    running an external command with the passed basename, implemented in terms
    of the Windows-specific Win32 API.

    See Also
    ----------
    https://stackoverflow.com/a/7110486/2809027
        StackOverflow answer strongly inspiring this implementation.
    :func:`is_parent_command`
        Further details.
    '''

    raise BetseFunctionUnimplementedException()



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

# ....................{ IMPORTS                           }....................
import matplotlib.pyplot as plt
import numpy as np
from betse.exceptions import BetseSimException, BetseSimConfException
from betse.lib.pickle import pickles
from betse.science import filehandling as fh
from betse.science.cells import Cells
from betse.science.chemistry.gene import MasterOfGenes
from betse.science.enum.enumconf import GrnUnpicklePhaseType
from betse.science.pipe.export.pipeexps import SimPipesExport
from betse.science.parameters import Parameters
from betse.science.phase import phasecallbacks
from betse.science.phase.phasecallbacks import SimCallbacksBCOrNoneTypes
from betse.science.phase.phasecls import SimPhase
from betse.science.enum.enumphase import SimPhaseKind
from betse.util.io.log import logs
from betse.util.path import files, pathnames
from betse.util.type.decorator.decorators import deprecated
from betse.util.type.decorator.decprof import log_time_seconds
from betse.util.type.text.string import strs
from betse.util.type.types import type_check
from matplotlib.collections import LineCollection, PolyCollection

# ....................{ CLASSES                           }....................
class SimRunner(object):
    '''
    **Simulation runner** (i.e., high-level object encapsulating the running of
    simulation phases as corresponding public methods commonly referred to as
    simulation subcommands).

    This runner provides high-level methods for initializing, running, and
    plotting simulations specified by the YAML-formatted simulation
    configuration file with which this runner is instantiated. Each instance of
    this runner simulates exactly one simulation.

    Attributes
    ----------
    _callbacks : SimCallbacksBC
        Caller-defined object whose methods are periodically called during each
        simulation subcommand (e.g., to notify this caller of phase progress).
    '''

    # ..................{ INITIALIZERS                      }..................
    @type_check
    def __init__(
        self,

        # Mandatory parameters.
        p: Parameters,

        # Optional parameters.
        callbacks: SimCallbacksBCOrNoneTypes = None,
    ) -> None:
        '''
        Initialize this simulation runner.

        Attributes
        ----------
        p : Parameters
            Simulation configuration to be run. If this configuration is in the
            unloaded state (i.e., has *not* yet been loaded into memory from a
            YAML-formatted file on disk), an exception is raised.
        callbacks : SimCallbacksBCOrNoneTypes
            Caller-defined object whose methods are periodically called during
            each simulation subcommand (e.g., :meth:`SimRunner.seed`). Defaults
            to ``None``, in which case this defaults to a placeholder object
            whose methods all silently reduce to noops.

        Raises
        ----------
        BetseYamlException
            If this simulation configuration is in the unloaded state.
        '''

        # If this configuration is unloaded, raise an exception.
        p.die_unless_loaded()

        # Default all unpassed parameters to sane defaults.
        if callbacks is None:
            callbacks = phasecallbacks.make_default()

        # Classify all passed parameters *AFTER* defaulting these parameters.
        self._callbacks = callbacks
        self._p = p

    # ..................{ RUNNERS ~ seed                    }..................
    @log_time_seconds(noun='seed')
    def seed(self) -> SimPhase:
        '''
        Seed this simulation with a new cell cluster and cache this cluster to
        an output file, specified by the current configuration file.

        This method *must* be called prior to the :meth:`init` and
        :meth:`plot_seed` methods, which consume this output as input.

        Returns
        ----------
        SimPhase
            High-level simulation phase instance encapsulating all objects
            internally created by this method to run this phase.
        '''

        # Log this attempt.
        logs.log_info('Seeding simulation...')

        # Cumulative number of times that each call of this subcommand calls
        # the SimCallbacksBC.progressed() callback or a callback calling that
        # callback (e.g., SimCallbacksBC.progressed_next()).
        #
        # This magic number *must* be manually synchronized with the
        # implementation of both this subcommand and methods transitively
        # called by this subcommand (e.g., Cells.make_world()). Failure to do
        # so *will* result in fatal exceptions. Sadly, there exists no
        # reasonable means of either automating or enforcing this constraint.
        SEED_PROGRESS_TOTAL = (
            # Number of progress callbacks performed directly in this method.
            6 +
            # Number of progress callbacks performed by Cells.make_world().
            Cells.MAKE_WORLD_PROGRESS_TOTAL
        )

        # Notify the caller of the range of work performed by this subcommand.
        self._callbacks.progress_ranged(progress_max=SEED_PROGRESS_TOTAL)

        # Simulation phase.
        phase = SimPhase(
            kind=SimPhaseKind.SEED, p=self._p, callbacks=self._callbacks)

        # Create the pseudo-randomized cell cluster.
        phase.cells.make_world(phase)

        # Initialize core simulation data structures.
        self._callbacks.progressed_next(
            status='Creating core computational matrices...')
        phase.sim.init_core(phase)

        # Define the tissue and boundary profiles for plotting.
        self._callbacks.progressed_next(
            status='Creating tissue, surgery, and boundary profiles...')
        phase.dyna.init_profiles(phase)

        # Redo gap junctions to isolate different tissue types.
        self._callbacks.progressed_next(
            status='Creating gap junction connection network...')
        phase.cells.redo_gj(phase)

        # Create a Laplacian and solver for discrete transfers on closed,
        # irregular cell network.
        self._callbacks.progressed_next(
            status='Creating cell network Poisson solver...')
        phase.cells.graphLaplacian(self._p)

        # Create accessory matrices depending on user requirements.
        if self._p.deformation:
            phase.cells.deform_tools(self._p)

        # if self._p.sim_eosmosis:
        #     phase.cells.eosmo_tools(self._p)

        # Pickle this cell cluster to disk.
        self._callbacks.progressed_next(
            status='Saving seeded cell cluster...')
        phase.cells.save_cluster(phase)

        # Log the completion of this phase.
        logs.log_info('Cell cluster creation complete!')
        phase.sim.sim_info_report(phase)

        # Signal the completion of this phase with respect to progress.
        self._callbacks.progressed_last()

        # Return this phase.
        return phase

    # ..................{ RUNNERS ~ (init|sim)              }..................
    @log_time_seconds(noun='initialization')
    def init(self) -> SimPhase:
        '''
        Initialize this simulation with the cell cluster seeded by a prior call
        to the :meth:`seed` method and cache this initialization to an output
        file, specified by the current configuration file.

        This method *must* be called prior to the :meth:`sim` and
        :meth:`plot_init` methods, which consume this output as input.

        Returns
        ----------
        SimPhase
            High-level simulation phase instance encapsulating all objects
            internally created by this method to run this phase.
        '''

        # Log this attempt.
        logs.log_info('Initializing simulation...')

        # Simulation phase type.
        phase_kind = SimPhaseKind.INIT

        if not files.is_file(self._p.seed_pickle_filename):
            if not self._p.autoInit:
                #FIXME: Call the _die_unless_file_pickled() function instead
                #both here and everywhere else we perform similar logic. Bugbears!
                raise BetseSimException(
                    'Initialization halted due to missing seed. '
                    'Please run "betse seed" and try again.')

            # Create an instance of world.
            logs.log_info('Automatically seeding cell cluster...')
            self.seed()
            logs.log_info('Now using seed to run initialization.')

        # Load the seed from cache.
        cells, p_old = fh.loadWorld(self._p.seed_pickle_filename)
        logs.log_info('Cell cluster loaded.')

        # check to ensure compatibility between original and present sim files:
        self._die_if_seed_differs(p_old, self._p)

        # Simulation phase, created *AFTER* unpickling these objects above.
        phase = SimPhase(
            kind=phase_kind,
            cells=cells,
            p=self._p,
            callbacks=self._callbacks,
        )

        # Initialize core simulation data structures.
        phase.sim.init_core(phase)

        # Run this simulation phase.
        phase.sim.sim_info_report(phase)
        phase.sim.run_sim_core(phase)

        # Return this phase.
        return phase


    @log_time_seconds(noun='simulation')
    def sim(self) -> SimPhase:
        '''
        Simulate this simulation with the cell cluster initialized by a prior
        call to the :meth:`init` method and cache this simulation to an output
        file, specified by the current configuration file.

        This method *must* be called prior to the :meth:`:meth:`plot_sim`
        method, which consumes this output as input.

        Returns
        ----------
        SimPhase
            High-level simulation phase instance encapsulating all objects
            internally created by this method to run this phase.
        '''

        # Log this attempt.
        logs.log_info('Running simulation...')

        # Simulation phase type.
        phase_kind = SimPhaseKind.SIM

        if not files.is_file(self._p.init_pickle_filename):
            if not self._p.autoInit:
                raise BetseSimException(
                    'Simulation halted due to missing initialization. '
                    'Please run "betse init" and try again.')

            logs.log_info('Automatically initializing cell cluster...')
            self.init()
            logs.log_info('Now using initialization to run simulation.')

        # Load the initialization from cache.
        sim, cells, p_old = fh.loadSim(self._p.init_pickle_filename)

        # Ensure compatibility between original and present config files.
        self._die_if_seed_differs(p_old, self._p)

        # Simulation phase, created *AFTER* unpickling these objects above.
        phase = SimPhase(
            kind=phase_kind,
            cells=cells,
            p=self._p,
            sim=sim,
            callbacks=self._callbacks,
        )

        # Run and save the simulation to the cache.
        sim.sim_info_report(phase)
        sim.run_sim_core(phase)

        # Return this phase.
        return phase

    # ..................{ RUNNERS ~ grn                     }..................
    @log_time_seconds(noun='network')
    def sim_grn(self) -> SimPhase:
        '''
        Initialize and simulate a pure gene regulatory network (GRN) *without*
        bioelectrics with the cell cluster seeded by a prior call to the
        :meth:`seed` method and cache this initialization and simulation to
        output files, specified by the current configuration file.

        This method *must* be called prior to the :meth:`plot_grn` method, which
        consumes this output as input.

        Returns
        ----------
        SimPhase
            High-level simulation phase instance encapsulating all objects
            internally created by this method to run this phase.
        '''

        # Simulation phase objects, defaulting to undefined initially.
        phase = None

        # Simulation phase type.
        phase_kind = SimPhaseKind.INIT

        # Simulator objects initialized below.
        cells = None
        sim = None

        # Log this simulation.
        logs.log_info(
            'Running gene regulatory network "%s" '
            'defined in config file "%s"...',
            pathnames.get_basename(self._p.grn_config_filename),
            self._p.conf_basename)

        # If networking an uninitialized, unsimulated cell cluster...
        if self._p.grn_unpickle_phase_type is GrnUnpicklePhaseType.SEED:
            if not files.is_file(self._p.seed_pickle_filename):
                if not self._p.autoInit:
                    raise BetseSimException(
                        'Simulation halted due to missing core seed. '
                        'Please run "betse seed" and try again.')

                # Create an instance of world.
                logs.log_info('Automatically seeding cell cluster...')
                self.seed()

            # Load the seed from cache.
            cells, _ = fh.loadWorld(self._p.seed_pickle_filename)
            logs.log_info('Running gene regulatory network on betse seed...')
            logs.log_info('Now using cell cluster to run initialization.')

            # Simulation phase.
            phase = SimPhase(
                kind=phase_kind,
                callbacks=self._callbacks,
                cells=cells,
                p=self._p,
            )

            # Initialize core simulation data structures.
            phase.sim.init_core(phase)
            phase.sim.init_dynamics(phase)

            #FIXME: Shift the following assignments into a new public
            #"Simulator" method -- say, Simulator.init_core_null().

            # Initialize other aspects required for piggyback of GRN on the
            # sim object.
            phase.sim.time = []
            phase.sim.vm = -50e-3 * np.ones(phase.sim.mdl)

            # Initialize key fields of simulator required to interface
            # (dummy init).
            phase.sim.rho_pump = 1.0
            phase.sim.rho_channel = 1.0
            phase.sim.conc_J_x = np.zeros(phase.sim.edl)
            phase.sim.conc_J_y = np.zeros(phase.sim.edl)
            phase.sim.J_env_x  = np.zeros(phase.sim.edl)
            phase.sim.J_env_y  = np.zeros(phase.sim.edl)
            phase.sim.u_env_x  = np.zeros(phase.sim.edl)
            phase.sim.u_env_y  = np.zeros(phase.sim.edl)
        # Else if networking an initialized but unsimulated cell cluster...
        elif self._p.grn_unpickle_phase_type is GrnUnpicklePhaseType.INIT:
            if not files.is_file(self._p.init_pickle_filename):
                if not self._p.autoInit:
                    raise BetseSimException(
                        'Simulation halted due to missing core initialization. '
                        'Please run "betse init" and try again.')

                logs.log_info('Automatically initializing cell cluster...')
                self.init()
                logs.log_info('Now using initialization to run simulation.')

            # Load the initialization from cache.
            logs.log_info('Running gene regulatory network on betse init...')
            sim, cells, _ = fh.loadSim(self._p.init_pickle_filename)
        # Else if networking an initialized, simulated cell cluster...
        elif self._p.grn_unpickle_phase_type is GrnUnpicklePhaseType.SIM:
            if not files.is_file(self._p.sim_pickle_filename):
                raise BetseSimException(
                    'Simulation halted due to missing core simulation. '
                    'Please run "betse sim" and try again.')

            # Load the simulation from cache.
            logs.log_info('Running gene regulatory network on betse sim...')
            sim, cells, _ = fh.loadSim(self._p.sim_pickle_filename)
        # Else, this type of networking is unrecognized. Raise an exception.
        else:
            raise BetseSimConfException(
                'Gene regulatory network (GRN) unpickle simulation phase '
                '"{}" unrecognized.'.format(self._p.grn_unpickle_phase_type))

        # If *NOT* defined above, define this simulation phase.
        if phase is None:
            phase = SimPhase(
                kind=phase_kind,
                callbacks=self._callbacks,
                cells=cells,
                p=self._p,
                sim=sim,
            )

            # Reinitialize all profiles.
            phase.dyna.init_profiles(phase)
            phase.dyna.init_events(phase)

        # If *NOT* restarting from a prior GRN run, start a new GRN.
        if self._p.grn_unpickle_filename is None:
            # Log this start.
            logs.log_info("Initializing the gene regulatory network...")

            # Create and initialize an instance of master of metabolism.
            MoG = MasterOfGenes(self._p)
            MoG.read_gene_config(phase)
        # Else, restart from a prior GRN run.
        else:
            # Log this restart.
            logs.log_info(
                'Reinitializing the gene regulatory network from "%s"...',
                pathnames.get_basename(self._p.grn_unpickle_filename))

            # If this file does *NOT* exist, raise an exception.
            _die_unless_file_pickled(
                filename=self._p.grn_unpickle_filename,
                subcommand='sim-grn',
                subcommand_label='Gene regulatory network')

            # Unpickle this file into a high-level "MasterOfGenes" object.
            MoG, _, _ = pickles.load(self._p.grn_unpickle_filename)

            # If running on a sim with a cut event, perform this cut...
            if (
                phase.dyna.event_cut is not None and
                phase.dyna.event_cut.is_fired
            ):
                # Log this cutting.
                logs.log_info(
                    'A cutting event has been run, '
                    'so the GRN object needs to be modified...')

                # If no prior initialization exists, raise an exception.
                if not files.is_file(self._p.sim_pickle_filename):
                    logs.log_warning(
                        'This situation is complex '
                        'due to a cutting event being run. '
                        'Please have a corresponding init file '
                        'to run the GRN simulation!')

                    raise BetseSimException(
                        'Simulation terminated due to missing core init. '
                        'Please alter GRN settings and try again.')
                # Else, a prior initialization exists.

                # Log this initialization.
                logs.log_info(
                    'Loading betse init from cache '
                    'for reference to original cells...')

                # Load the initialization from cache.
                sim_old, cells_old, _ = fh.loadSim(self._p.sim_pickle_filename)

                #FIXME: This phase object would ideally be pickled to and
                #from the "self._p.sim_pickle_filename" file loaded above, in
                #which case this local variable would be safely removable.

                # Original simulation phase. To avoid caller confusion, the
                # optional "callbacks" parameter is intentionally *NOT* passed.
                phase_old = SimPhase(
                    kind=phase_kind,
                    p=self._p,
                    cells=cells_old,
                    sim=sim_old,
                )

                # Initialize all tissue profiles on original cells.
                phase_old.dyna.init_profiles(phase_old)

                for cut_profile_name in phase_old.p.event_cut_profile_names:
                    logs.log_info(
                        'Cutting cell cluster via cut profile "%s"...',
                        cut_profile_name)

                    # Object picking the cells removed by this cut profile.
                    tissue_picker = phase_old.dyna.cut_name_to_profile[
                        cut_profile_name].picker

                    # One-dimensional Numpy arrays of the indices of all
                    # cells and cell membranes to be removed.
                    target_inds_cell, target_inds_mems = (
                        tissue_picker.pick_cells_and_mems(
                            cells=cells_old, p=self._p))

                    MoG.core.mod_after_cut_event(
                        phase, target_inds_cell, target_inds_mems)
                    MoG.core.redefine_dynamic_dics(sim, cells, self._p)

                    logs.log_info(
                        'Reinitializing gene regulatory network '
                        'for simulation...')
                    MoG.reinitialize(phase)

            # if self._p.use_microtubules:
            #     sim.mtubes.mtubes_x = MoG.mtubes_x_time[-1]
            #     sim.mtubes.mtubes_y = MoG.mtubes_y_time[-1]
            #
            #     sim.mtubes.uxmt, sim.mtubes.uymt = sim.mtubes.mtubes_to_cell(
            #         cells, self._p)

        logs.log_info('Simulating gene regulatory network...')
        MoG.run_core_sim(phase)

        # Return this phase.
        return phase

    # ..................{ PLOTTERS                          }..................
    @log_time_seconds(noun='seed', verb='exported')
    def plot_seed(self) -> SimPhase:
        '''
        Visualize the cell cluster seed by a prior call to the :meth:`seed`
        method and export the resulting plots and animations to various output
        files, specified by the current configuration file.

        Returns
        ----------
        SimPhase
            High-level simulation phase instance encapsulating all objects
            internally created by this method to run this phase.
        '''

        # Log this plotting attempt.
        logs.log_info(
            'Plotting cell cluster with configuration file "%s".',
            self._p.conf_basename)

        # If an initialization does *NOT* already exist, raise an exception.
        _die_unless_file_pickled(
            filename=self._p.seed_pickle_filename,
            subcommand='seed',
            subcommand_label='Seed')

        # Load the seed from cache.
        cells, _ = fh.loadWorld(self._p.seed_pickle_filename)
        logs.log_info('Cell cluster loaded.')

        # Simulation phase, created *AFTER* unpickling these objects above
        phase = SimPhase(
            kind=SimPhaseKind.SEED,
            p=self._p,
            cells=cells,
            callbacks=self._callbacks,
        )

        # Initialize core simulation data structures.
        phase.sim.init_core(phase)
        phase.dyna.init_profiles(phase)

        #FIXME: Refactor into a seed-specific plot pipeline. Dreaming androids!
        if self._p.autosave:
            savedImg = pathnames.join(self._p.init_export_dirname, 'fig_')

        # if self._p.plot_cell_cluster:
        # fig_tiss, ax_tiss, cb_tiss = viz.clusterPlot(
        #     self._p, phase.dyna, cells, clrmap=self._p.background_cm)

        if self._p.autosave:
            savename10 = savedImg + 'cluster_mosaic' + '.png'
            plt.savefig(savename10, format='png', transparent=True)

        if self._p.plot.is_after_sim_show:
            plt.show(block=False)

        if self._p.is_ecm:  # and self._p.plot_cluster_mask:
            plt.figure()
            ax99 = plt.subplot(111)
            plt.imshow(
                np.log10(phase.sim.D_env_weight.reshape(phase.cells.X.shape)),
                origin='lower',
                extent=[
                    self._p.um * phase.cells.xmin,
                    self._p.um * phase.cells.xmax,
                    self._p.um * phase.cells.ymin,
                    self._p.um * phase.cells.ymax,
                ],
                cmap=self._p.background_cm,
            )
            plt.colorbar()

            cell_edges_flat = self._p.um * phase.cells.mem_edges_flat
            coll = LineCollection(cell_edges_flat, colors='k')
            coll.set_alpha(1.0)
            ax99.add_collection(coll)

            plt.title('Logarithm of Environmental Diffusion Weight Matrix')

            if self._p.autosave:
                savename10 = savedImg + 'env_diffusion_weights' + '.png'
                plt.savefig(savename10, format='png', transparent=True)

            if self._p.plot.is_after_sim_show:
                plt.show(block=False)

            plt.figure()
            plt.imshow(
                cells.maskM,
                origin='lower',
                extent=[
                    self._p.um * phase.cells.xmin,
                    self._p.um * phase.cells.xmax,
                    self._p.um * phase.cells.ymin,
                    self._p.um * phase.cells.ymax,
                ],
                cmap=self._p.background_cm,
            )
            plt.colorbar()
            plt.title('Cluster Masking Matrix')

            if self._p.autosave:
                savename = savedImg + 'cluster_mask' + '.png'
                plt.savefig(savename, format='png', transparent=True)

            if self._p.plot.is_after_sim_show:
                plt.show(block=False)

        # Plot gap junctions.
        # if self._p.plot_cell_connectivity:
        plt.figure()
        ax_x = plt.subplot(111)

        if self._p.showCells:
            base_points = np.multiply(phase.cells.cell_verts, self._p.um)
            col_cells = PolyCollection(
                base_points, facecolors='k', edgecolors='none')
            col_cells.set_alpha(0.3)
            ax_x.add_collection(col_cells)

        con_segs = phase.cells.nn_edges
        connects = self._p.um * np.asarray(con_segs)
        collection = LineCollection(connects, linewidths=1.0, color='b')
        ax_x.add_collection(collection)
        plt.axis('equal')
        plt.axis([
            self._p.um * phase.cells.xmin,
            self._p.um * phase.cells.xmax,
            self._p.um * phase.cells.ymin,
            self._p.um * phase.cells.ymax,
        ])

        ax_x.set_xlabel('Spatial x [um]')
        ax_x.set_ylabel('Spatial y [um')
        ax_x.set_title('Cell Connectivity Network')

        if self._p.autosave:
            savename10 = savedImg + 'gj_connectivity_network' + '.png'
            plt.savefig(savename10, format='png', transparent=True)

        if self._p.turn_all_plots_off is False:
            plt.show(block=False)
        else:
            logs.log_info(
                'Plots exported to init results folder '
                'defined in configuration file "%s".',
                self._p.conf_basename)

        # Return this phase.
        return phase


    @log_time_seconds(noun='initialization', verb='exported')
    def plot_init(self) -> SimPhase:
        '''
        Visualize the cell cluster initialized by a prior call to the
        :meth:`init` method and export the resulting plots and animations to
        various output files, specified by the current configuration file.

        Returns
        ----------
        SimPhase
            High-level simulation phase instance encapsulating all objects
            internally created by this method to run this phase.
        '''

        # Log this plotting attempt.
        logs.log_info(
            'Plotting initialization with configuration "%s"...',
            self._p.conf_basename)

        # If an initialization does *NOT* already exist, raise an exception.
        _die_unless_file_pickled(
            filename=self._p.init_pickle_filename,
            subcommand='init',
            subcommand_label='Initialization')

        # Simulation phase type.
        phase_kind = SimPhaseKind.INIT

        # Load the initialization from cache.
        sim, cells, _ = fh.loadSim(self._p.init_pickle_filename)

        # Simulation phase, created *AFTER* unpickling these objects above
        phase = SimPhase(
            kind=phase_kind,
            cells=cells,
            p=self._p,
            sim=sim,
            callbacks=self._callbacks,
        )

        #FIXME: This... isn't the best. Ideally, the phase.dyna.init_profiles()
        #method would *ALWAYS* be implicitly called by the SimPhase.__init__()
        #method. Unfortunately, the non-trivial complexity of cell cluster
        #initialization requires we do so manually for now. Sad sandlion frowns!
        phase.dyna.init_profiles(phase)

        # Display and/or save all initialization exports (e.g., animations).
        SimPipesExport().export(phase)

        #FIXME: All of the following crash if image saving is not turned on, but
        #due to whatever way this is set up, it's not possible to readily fix
        #it. Grrrrrr.....
        #FIXME: Why not just stop each block from happening if image saving is
        #off? Easy peasy!

        #FIXME: Reduce duplication. The following logic is effectively a copy of
        #similar logic in the plot_sim() method.
        #FIXME: Split each of the following blocks performing both plotting and
        #animating into their appropriate plotpipe.pipeline() or
        #animpipe.pipeline() functions, which should resolve the above concerns.

        # run the molecules plots:
        if self._p.molecules_enabled and sim.molecules is not None:
            # reinit settings for plots, in case they've changed:
            sim.molecules.core.plot_init(self._p.network_config, self._p)
            sim.molecules.core.init_saving(cells, self._p, plot_type='init')
            sim.molecules.core.export_all_data(sim, cells, self._p, message='auxiliary molecules')
            sim.molecules.core.plot(sim, cells, self._p, message='auxiliary molecules')
            sim.molecules.core.anim(phase=phase, message='auxiliary molecules')

        if self._p.grn_enabled and sim.grn is not None:
            # reinitialize the plot settings:
            sim.grn.core.plot_init(self._p.grn.conf, self._p)
            sim.grn.core.init_saving(
                cells, self._p, plot_type='init', nested_folder_name='GRN')
            sim.grn.core.export_all_data(sim, cells, self._p, message='GRN molecules')
            sim.grn.core.plot(sim, cells, self._p, message='GRN molecules')
            sim.grn.core.anim(phase=phase, message='GRN molecules')

        # If displaying plots, block on all previously plots previously
        # displayed as non-blocking. If this is *NOT* done, these plots will
        # effectively *NEVER* displayed be on most systems.
        if self._p.plot.is_after_sim_show:
            plt.show()

        # Return this phase.
        return phase


    @log_time_seconds(noun='simulation', verb='exported')
    def plot_sim(self) -> SimPhase:
        '''
        Visualize the cell cluster simulated by a prior call to the :meth:`sim`
        method and export the resulting plots and animations to various output
        files, specified by the current configuration file.

        Returns
        ----------
        SimPhase
            High-level simulation phase instance encapsulating all objects
            internally created by this method to run this phase.
        '''

        # Log this plotting attempt.
        logs.log_info(
            'Plotting simulation with configuration "%s"...',
            self._p.conf_basename)

        # If a simulation does *NOT* already exist, raise an exception.
        _die_unless_file_pickled(
            filename=self._p.sim_pickle_filename,
            subcommand='sim',
            subcommand_label='Simulation')

        # Simulation phase type.
        phase_kind = SimPhaseKind.SIM

        # Load the simulation from the cache.
        sim, cells, _ = fh.loadSim(self._p.sim_pickle_filename)

        # Simulation phase, created *AFTER* unpickling these objects above
        phase = SimPhase(
            kind=phase_kind,
            cells=cells,
            p=self._p,
            sim=sim,
            callbacks=self._callbacks,
        )

        #FIXME: This... isn't the best. Ideally, the phase.dyna.init_profiles()
        #method would *ALWAYS* be implicitly called by the SimPhase.__init__()
        #method. Unfortunately, the non-trivial complexity of cell cluster
        #initialization requires we do so manually for now. Sad sandlion frowns!
        phase.dyna.init_profiles(phase)

        # Display and/or save all simulation exports (e.g., animations).
        SimPipesExport().export(phase)

        #FIXME: Split each of the following blocks performing both plotting and
        #animating into their appropriate plotpipe.pipeline() or
        #animpipe.pipeline() functions.

        # run the molecules plots:
        if self._p.molecules_enabled and sim.molecules is not None:
            # reinit settings for plots, in case they've changed:
            sim.molecules.core.plot_init(self._p.network_config, self._p)
            sim.molecules.core.init_saving(cells, self._p, plot_type='sim')
            sim.molecules.core.export_all_data(sim, cells, self._p, message='auxiliary molecules')
            sim.molecules.core.plot(sim, cells, self._p, message='auxiliary molecules')
            sim.molecules.core.anim(phase=phase, message='auxiliary molecules')

        if self._p.grn_enabled and sim.grn is not None:
            # reinitialize the plot settings:
            sim.grn.core.plot_init(self._p.grn.conf, self._p)
            sim.grn.core.init_saving(cells, self._p, plot_type='sim', nested_folder_name='GRN')
            sim.grn.core.export_all_data(sim, cells, self._p, message='GRN molecules')
            sim.grn.core.plot(sim, cells, self._p, message='GRN molecules')
            sim.grn.core.anim(phase=phase, message='GRN molecules')

        # If displaying plots, block on all previously plots previously
        # displayed as non-blocking. If this is *NOT* done, these plots will
        # effectively *NEVER* displayed be on most systems.
        if self._p.plot.is_after_sim_show:
            plt.show()

        # Return this phase.
        return phase


    def plot_grn(self) -> SimPhase:
        '''
        Visualize the pure gene regulatory network (GRN) initialized and
        simulated by a prior call to the :meth:`sim_grn` method and export the
        resulting plots and animations to various output files, specified by the
        current configuration file.

        Returns
        ----------
        SimPhase
            High-level simulation phase instance encapsulating all objects
            internally created by this method to run this phase.
        '''

        # Simulation phase type.
        phase_kind = SimPhaseKind.INIT

        # MoG = MasterOfGenes(self._p)
        MoG, cells, _ = fh.loadSim(self._p.grn_pickle_filename)

        # Simulation phase.
        phase = SimPhase(
            kind=phase_kind,
            cells=cells,
            p=self._p,
            callbacks=self._callbacks,
        )

        # Initialize core simulation data structures.
        phase.sim.init_core(phase)

        # Initialize simulation data structures
        # sim.baseInit_all(cells, p)

        #FIXME: This... looks a bit hacky. That said, maybe it is good? Thunder
        #dragons unite!
        phase.sim.time = MoG.time

        MoG.core.plot_init(self._p.grn.conf, self._p)
        MoG.core.init_saving(
            cells, self._p, plot_type='grn', nested_folder_name='RESULTS')
        MoG.core.export_all_data(
            phase.sim, cells, self._p, message='gene products')
        MoG.core.plot(
            phase.sim, cells, self._p, message='gene products')
        MoG.core.anim(phase=phase, message='gene products')

        # If displaying plots, block on all previously plots previously
        # displayed as non-blocking. If this is *NOT* done, these plots will
        # effectively *NEVER* displayed be on most systems.
        if self._p.plot.is_after_sim_show:
            plt.show()

        # Return this phase.
        return phase

    # ..................{ EXCEPTIONS                         }..................
    @type_check
    def _die_if_seed_differs(
        self,
        p_old: Parameters,
        p_new: Parameters,
    ) -> None:
        '''
        Raise an exception if one or more general or seed (i.e., world) options
        differ between the passed simulation configurations, implying the
        current configuration to have been unsafely modified since the initial
        creation of the cell cluster for this configuration.

        Attributes
        ----------
        p_old : Parameters
            Previous simulation configuration to be compared.
        p_new : Parameters
            Current simulation configuration to be compared.

        Raises
        ----------
        :class:`BetseSimConfException`
            If these options differ for these configurations.
        '''

        if (p_old.config['general options'] != p_new.config['general options'] or
            p_old.config['world options'  ] != p_new.config['world options']):
            raise BetseSimConfException(
                'Important config file options are out of sync between '
                'seed and this init/sim attempt! '
                'Run "betse seed" again to match the current settings of '
                'this config file.')

    # ..................{ DEPRECATED                        }..................
    # The following methods have been deprecated for compliance with PEP 8.

    #FIXME: Remove all deprecated methods defined below *AFTER* a sufficient
    #amount of time -- say, mid to late 2018.

    @deprecated
    def makeWorld(self) -> None:
        return self.seed()

    @deprecated
    def initialize(self) -> None:
        return self.init()

    @deprecated
    def simulate(self) -> None:
        return self.sim()

    @deprecated
    def plotInit(self) -> None:
        return self.plot_init()

    @deprecated
    def plotSim(self) -> None:
        return self.plot_sim()

# ....................{ EXCEPTIONS                        }....................
@type_check
def _die_unless_file_pickled(
    filename: str, subcommand: str, subcommand_label: str) -> None:
    '''
    Raise an exception containing the passed machine-readable subcommand and
    human-readable label denoting that subcommand unless the file with the
    passed filename pickled by a prior run of that subcommand exists.

    Parameters
    ----------
    filename : str
        Absolute filename of the pickled file to be validated.
    subcommand : str
        Machine-readable case-sensitive name of the prior subcommand expected
        to have pickled this file (e.g., ``seed``, ``init``).
    subcommand_label : str
        Human-readable case-insensitive noun denoting the type of prior
        subcommand expected to have pickled this file (e.g., ``Seed``,
        ``initialization``).

    Raises
    ----------
    BetseSimException
        If this file does *not* exist.
    '''

    # If this file does *NOT* exist...
    if not files.is_file(filename):
        # Uppercase the first character of this label for readability.
        subcommand_label_cased = strs.uppercase_char_first(subcommand_label)

        # Raise an exception embedding these parameters.
        raise BetseSimException(
            '{} not previously run; '
            'please run "betse {}" and try again.\n'
            'Specifically, file not found: {}'.format(
                subcommand_label_cased, subcommand, filename))
